# Generated by Django 4.2.23 on 2025-11-20 21:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0082_remove_bigbrotherconfig_token_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='bigbrotherredditmessage',
            options={'ordering': ['-created'], 'verbose_name': 'Reddit Recruitment Message', 'verbose_name_plural': 'Reddit Recruitment Messages'},
        ),
        migrations.AlterModelOptions(
            name='bigbrotherredditsettings',
            options={'verbose_name': 'BigBrother Reddit Settings'},
        ),
        migrations.AlterModelOptions(
            name='complianceticket',
            options={'ordering': ['-created_at'], 'verbose_name': 'Compliance Ticket', 'verbose_name_plural': 'Compliance Tickets'},
        ),
        migrations.AlterModelOptions(
            name='leaverequest',
            options={'ordering': ['-created_at'], 'verbose_name': 'Leave of Absence Request', 'verbose_name_plural': 'Leave of Absence Requests'},
        ),
        migrations.AlterModelOptions(
            name='optmessages1',
            options={'verbose_name': 'Optional Message 1', 'verbose_name_plural': 'Optional Messages 1'},
        ),
        migrations.AlterModelOptions(
            name='optmessages2',
            options={'verbose_name': 'Optional Message 2', 'verbose_name_plural': 'Optional Messages 2'},
        ),
        migrations.AlterModelOptions(
            name='optmessages3',
            options={'verbose_name': 'Optional Message 3', 'verbose_name_plural': 'Optional Messages 3'},
        ),
        migrations.AlterModelOptions(
            name='optmessages4',
            options={'verbose_name': 'Optional Message 4', 'verbose_name_plural': 'Optional Messages 4'},
        ),
        migrations.AlterModelOptions(
            name='optmessages5',
            options={'verbose_name': 'Optional Message 5', 'verbose_name_plural': 'Optional Messages 5'},
        ),
        migrations.AlterModelOptions(
            name='papcompliance',
            options={'verbose_name': 'PAP Compliance Score', 'verbose_name_plural': 'PAP Compliance Scores'},
        ),
        migrations.AlterModelOptions(
            name='tickettoolconfig',
            options={'verbose_name': 'Ticket Tool Configuration', 'verbose_name_plural': 'Ticket Tool Configuration'},
        ),
        migrations.AlterModelOptions(
            name='warmprogress',
            options={'verbose_name': 'Warm Preload Progress', 'verbose_name_plural': 'Warm Preload Progress'},
        ),
        migrations.RenameField(
            model_name='userstatus',
            old_name='has_imp_blacklist',
            new_name='has_alliance_blacklist',
        ),
        migrations.RenameField(
            model_name='userstatus',
            old_name='has_lawn_blacklist',
            new_name='has_coalition_blacklist',
        ),
    ]



# Generated by Django 4.2.23 on 2025-08-17 18:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0046_bigbrotherconfig_member_corporations'),
    ]

    operations = [
        migrations.AddField(
            model_name='bigbrotherconfig',
            name='member_alliances',
            field=models.TextField(blank=True, help_text="List of alliance IDs to be considered members, separated by ','", null=True),
        ),
    ]



# Generated by Django 4.2.21 on 2025-08-03 12:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0018_userstatus_cyno_userstatus_has_skills_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='bigbrotherconfig',
            name='whitelist_alliances',
            field=models.TextField(blank=True, default='', help_text="List of alliance IDs considered whitelisted, separated by ','", null=True),
        ),
        migrations.AddField(
            model_name='bigbrotherconfig',
            name='whitelist_corporations',
            field=models.TextField(blank=True, help_text="List of corporation IDs considered whitelisted, separated by ','", null=True),
        ),
    ]



from django.test import TestCase, Client
from django.urls import reverse
from django.contrib.auth.models import User, Permission
from django.contrib.contenttypes.models import ContentType
from allianceauth.authentication.models import UserProfile, State, CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from aa_bb.models import BigBrotherConfig, General
from aa_bb.views import get_available_cards, CARD_DEFINITIONS
import json
from unittest.mock import patch

class TestDashboardViews(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Create states with unique names and likely unique priorities
        cls.member_state, _ = State.objects.get_or_create(name="BB_Member", defaults={'priority': 10001})
        cls.guest_state, _ = State.objects.get_or_create(name="BB_Guest", defaults={'priority': 10002})

        # Create users with passwords
        cls.admin_user = User.objects.create_user(username="admin", password="password")
        cls.recruiter_user = User.objects.create_user(username="recruiter", password="password")
        cls.basic_user = User.objects.create_user(username="basic", password="password")
        cls.super_user = User.objects.create_superuser(username="superuser", password="password", email="admin@example.com")

        # Add permissions
        content_type = ContentType.objects.get_for_model(General)
        perms = Permission.objects.filter(content_type=content_type)

        basic_access = perms.get(codename="basic_access")
        full_access = perms.get(codename="full_access")
        recruiter_access = perms.get(codename="recruiter_access")

        cls.admin_user.user_permissions.add(basic_access, full_access)
        cls.recruiter_user.user_permissions.add(basic_access, recruiter_access)
        cls.basic_user.user_permissions.add(basic_access)

        # Refresh users to ensure permissions are loaded
        cls.admin_user = User.objects.get(pk=cls.admin_user.pk)
        cls.recruiter_user = User.objects.get(pk=cls.recruiter_user.pk)
        cls.basic_user = User.objects.get(pk=cls.basic_user.pk)

        # Create config
        cls.config = BigBrotherConfig.get_solo()
        cls.config.is_active = True
        cls.config.bb_member_states.add(cls.member_state)
        cls.config.bb_guest_states.add(cls.guest_state)
        cls.config.alliance_blacklist_url = "http://alliance.bl"
        cls.config.external_blacklist_url = "http://external.bl"
        cls.config.save()

        # Create characters and ownership
        # We give the "viewer" users the Member state so they don't show up in Guest lists
        cls.char_admin = EveCharacter.objects.create(character_id=100, character_name="Admin Char", corporation_id=1, corporation_name="Admin Corp")
        CharacterOwnership.objects.create(character=cls.char_admin, user=cls.admin_user, owner_hash="h0")
        cls.admin_user.profile.main_character = cls.char_admin
        cls.admin_user.profile.state = cls.member_state
        cls.admin_user.profile.save()
        UserProfile.objects.filter(pk=cls.admin_user.profile.pk).update(state=cls.member_state)

        cls.char_recruiter = EveCharacter.objects.create(character_id=101, character_name="Recruiter Char", corporation_id=1, corporation_name="Member Corp")
        CharacterOwnership.objects.create(character=cls.char_recruiter, user=cls.recruiter_user, owner_hash="h11")
        cls.recruiter_user.profile.main_character = cls.char_recruiter
        cls.recruiter_user.profile.state = cls.member_state
        cls.recruiter_user.profile.save()
        UserProfile.objects.filter(pk=cls.recruiter_user.profile.pk).update(state=cls.member_state)

        cls.char_basic = EveCharacter.objects.create(character_id=102, character_name="Basic Char", corporation_id=1, corporation_name="Member Corp")
        CharacterOwnership.objects.create(character=cls.char_basic, user=cls.basic_user, owner_hash="h12")
        cls.basic_user.profile.main_character = cls.char_basic
        cls.basic_user.profile.state = cls.member_state
        cls.basic_user.profile.save()
        UserProfile.objects.filter(pk=cls.basic_user.profile.pk).update(state=cls.member_state)

        cls.char_member = EveCharacter.objects.create(character_id=103, character_name="Member Char", corporation_id=1, corporation_name="Member Corp")
        cls.user_member = User.objects.create_user(username="member_user", password="password")
        CharacterOwnership.objects.create(character=cls.char_member, user=cls.user_member, owner_hash="h1")
        cls.user_member.profile.state = cls.member_state
        cls.user_member.profile.main_character = cls.char_member
        cls.user_member.profile.save()
        UserProfile.objects.filter(pk=cls.user_member.profile.pk).update(state=cls.member_state)

        cls.char_guest = EveCharacter.objects.create(character_id=104, character_name="Guest Char", corporation_id=2, corporation_name="Guest Corp")
        cls.user_guest = User.objects.create_user(username="guest_user", password="password")
        CharacterOwnership.objects.create(character=cls.char_guest, user=cls.user_guest, owner_hash="h2")
        cls.user_guest.profile.state = cls.guest_state
        cls.user_guest.profile.main_character = cls.char_guest
        cls.user_guest.profile.save()
        UserProfile.objects.filter(pk=cls.user_guest.profile.pk).update(state=cls.guest_state)

        # Superuser also needs a profile/main character to avoid redirects
        cls.char_super = EveCharacter.objects.create(character_id=105, character_name="Super Char", corporation_id=1, corporation_name="Admin Corp")
        CharacterOwnership.objects.create(character=cls.char_super, user=cls.super_user, owner_hash="h99")
        cls.super_user.profile.main_character = cls.char_super
        cls.super_user.profile.state = cls.member_state
        cls.super_user.profile.save()
        UserProfile.objects.filter(pk=cls.super_user.profile.pk).update(state=cls.member_state)

    def setUp(self):
        self.client = Client()

    def test_index_access(self):
        # Anonymous users should be redirected to login
        response = self.client.get(reverse("aa_bb:index"))
        self.assertEqual(response.status_code, 302)

        # Authorized users should get 200
        self.assertTrue(self.client.login(username="basic", password="password"))
        response = self.client.get(reverse("aa_bb:index"))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "aa_bb/index.html")

    def test_index_dropdown_options_full_access(self):
        self.assertTrue(self.client.login(username="admin", password="password"))
        response = self.client.get(reverse("aa_bb:index"))
        self.assertEqual(response.status_code, 200)
        options = response.context['dropdown_options']
        self.assertIn("Member Char", options)
        self.assertIn("Guest Char", options)

    def test_superuser_access(self):
        self.assertTrue(self.client.login(username="superuser", password="password"))
        response = self.client.get(reverse("aa_bb:index"))
        self.assertEqual(response.status_code, 200)

    def test_index_dropdown_options_recruiter_access(self):
        self.assertTrue(self.client.login(username="recruiter", password="password"))
        response = self.client.get(reverse("aa_bb:index"))
        self.assertEqual(response.status_code, 200)
        options = response.context['dropdown_options']
        self.assertNotIn("Member Char", options)
        self.assertIn("Guest Char", options)

    def test_available_cards_logic(self):
        # With URLs configured, both blacklist cards should be present
        cards = get_available_cards()
        keys = [c['key'] for c in cards]
        self.assertIn('alliance_bl', keys)
        self.assertIn('external_bl', keys)

        # Clear URLs and verify they are removed
        config = BigBrotherConfig.get_solo()
        config.alliance_blacklist_url = ""
        config.external_blacklist_url = ""
        config.save()

        cards = get_available_cards()
        keys = [c['key'] for c in cards]
        self.assertNotIn('alliance_bl', keys)
        self.assertNotIn('external_bl', keys)

        # Restore for other tests
        config.alliance_blacklist_url = "http://alliance.bl"
        config.external_blacklist_url = "http://external.bl"
        config.save()

    @patch('aa_bb.views.get_card_data')
    def test_load_card_ajax(self, mock_get_card_data):
        mock_get_card_data.return_value = ("<p>Test Card Content</p>", True)
        self.assertTrue(self.client.login(username="admin", password="password"))

        # Find index of a standard card (e.g., 'skills')
        cards = get_available_cards()
        skills_idx = next(i for i, c in enumerate(cards) if c['key'] == 'skills')

        url = reverse("aa_bb:load_card")
        response = self.client.get(url, {'option': 'Member Char', 'index': skills_idx})

        self.assertEqual(response.status_code, 200)
        data = response.json()
        self.assertEqual(data['title'], 'Skills')
        self.assertEqual(data['content'], "<p>Test Card Content</p>")
        self.assertTrue(data['status'])

    def test_load_card_invalid_params(self):
        self.assertTrue(self.client.login(username="admin", password="password"))
        url = reverse("aa_bb:load_card")

        # Missing parameters
        response = self.client.get(url)
        self.assertEqual(response.status_code, 400)

        # Invalid index
        response = self.client.get(url, {'option': 'Member Char', 'index': 999})
        self.assertEqual(response.status_code, 400)

        # Unknown character
        response = self.client.get(url, {'option': 'Nonexistent Char', 'index': 0})
        self.assertEqual(response.status_code, 404)

    def test_manual_modules_view(self):
        self.assertTrue(self.client.login(username="admin", password="password"))
        url = reverse("aa_bb:manual_modules")
        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "faq/modules.html")
        self.assertIn('modules', response.context)

        # Verify some module names are present
        content = response.content.decode()
        self.assertIn("BigBrother Core Dashboard", content)
        self.assertIn("CorpBrother Dashboard", content)




# Generated by Django 4.2.23 on 2026-01-03 19:42

from django.db import migrations, models
from django.conf import settings
import django.db.models.deletion

class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0100_tickettoolconfig_use_forum_threads'),
    ]

    operations = [
        migrations.AddField(
            model_name='userstatus',
            name='compliance_forum_thread_id',
            field=models.BigIntegerField(blank=True, null=True),
        ),
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False,
                     'permissions': (('basic_access', 'Can access Big Brother'),
                                     ('full_access', 'Can view all main characters in Big Brother'),
                                     ('recruiter_access',
                                      'Can view main characters in Guest state only in Big Brother'),
                                     ('basic_access_cb', 'Can access Corp Brother'),
                                     ('full_access_cb', 'Can view all corps in Corp Brother'),
                                     ('recruiter_access_cb', "Can view guest's corps only in Corp Brother"),
                                     ('can_blacklist_characters', 'Can add characters to blacklist'),
                                     ('can_access_loa', 'Can access and submit a Leave Of Absence request'),
                                     ('can_view_all_loa', 'Can view all Leave Of Absence requests'),
                                     ('can_manage_loa', 'Can manage Leave Of Absence requests'),
                                     ('can_access_paps', 'Can access PAP Stats'),
                                     ('can_generate_paps', 'Can generate PAP Stats'),
                                     ('ticket_manager', 'Can manage compliance tickets'))},
        ),
        migrations.CreateModel(
            name='ComplianceTicketComment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('comment', models.TextField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('ticket', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='comments',
                                             to='aa_bb.complianceticket')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL,
                                           to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Ticket Comment',
                'verbose_name_plural': 'Ticket Comments',
                'ordering': ['created_at'],
            },
        ),
        migrations.AddField(
            model_name='tickettoolconfig',
            name='Forum_Channel_ID',
            field=models.PositiveBigIntegerField(blank=True, default=0,
                                                 help_text='Channel ID to create threads in (for Bot-managed threads)',
                                                 null=True, verbose_name='Forum/Thread Channel ID'),
        ),
        migrations.AddField(
            model_name='tickettoolconfig',
            name='ticket_type',
            field=models.CharField(
                choices=[('private_channel', 'Private Channels (Bot)'), ('private_thread', 'Private Threads (Bot)'),
                         ('forum_thread', 'Public Forum Threads (Webhook)'), ('auth_only', 'Auth Only (No Discord)')],
                default='private_channel',
                help_text="Choose how compliance tickets are created. 'Private Threads' requires aadiscordbot.",
                max_length=20, verbose_name='Ticket Type'),
        ),
        migrations.AlterField(
            model_name='complianceticket',
            name='reason',
            field=models.CharField(choices=[('corp_check', 'Corp Compliance'), ('paps_check', 'PAP Requirements'),
                                            ('afk_check', 'Inactivity'), ('discord_check', 'User is not on discord'),
                                            ('char_removed', 'Character removed'), ('awox_kill', 'AWOX kill found')],
                                   max_length=20),
        ),
        migrations.CreateModel(
            name='ComplianceThread',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('reason', models.CharField(max_length=20)),
                ('thread_id', models.BigIntegerField()),
                ('user',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='compliance_threads',
                                   to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Compliance Thread',
                'verbose_name_plural': 'Compliance Threads',
                'unique_together': {('user', 'reason')},
            },
        ),
        migrations.AlterField(
            model_name='complianceticket',
            name='discord_channel_id',
            field=models.BigIntegerField(blank=True, db_index=True, null=True),
        ),
        migrations.AddField(
            model_name='tickettoolconfig',
            name='awox_kill_reason',
            field=models.TextField(blank=True,
                                   default='<@&{role}>,<@{namee}> detection indicates your involvement in an AWOX kill, please explain:\n{details}',
                                   help_text='Message to send with {role}, {namee} and {details} variables', null=True),
        ),
        migrations.AddField(
            model_name='tickettoolconfig',
            name='char_removed_reason',
            field=models.TextField(blank=True,
                                   default='<@&{role}>,<@{namee}> Auth lost access to your character {details}, please fix it ASAP.',
                                   help_text='Message to send with {role}, {namee} and {details} variables', null=True),
        ),
    ]



# Generated by Django 4.2.23 on 2025-08-06 09:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0030_alter_general_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='bigbrotherconfig',
            name='loa_max_logoff_days',
            field=models.IntegerField(default=30, help_text='How many days can a user not login w/o a loa request before notifications'),
        ),
        migrations.AlterField(
            model_name='bigbrotherconfig',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='entityinfocache',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='leaverequest',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='messages',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages1',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages2',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages3',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages4',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages5',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='suscontractnote',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='susmailnote',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='sustransactionnote',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='userstatus',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='warmprogress',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



# Generated by Django 4.2.20 on 2025-05-02 08:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0013_bigbrotherconfig_mail_keywords'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProcessedContract',
            fields=[
                ('contract_id', models.BigIntegerField(help_text="The Contract.contract_id that we've processed", primary_key=True, serialize=False)),
                ('processed_at', models.DateTimeField(auto_now_add=True, help_text='When we first processed this contract')),
            ],
            options={
                'verbose_name': 'Processed Contract',
                'verbose_name_plural': 'Processed Contracts',
                'db_table': 'aa_bb_processed_contracts',
            },
        ),
        migrations.CreateModel(
            name='ProcessedMail',
            fields=[
                ('mail_id', models.BigIntegerField(help_text="The MailMessage.id_key that we've processed", primary_key=True, serialize=False)),
                ('processed_at', models.DateTimeField(auto_now_add=True, help_text='When we first processed this mail')),
            ],
            options={
                'verbose_name': 'Processed Mail',
                'verbose_name_plural': 'Processed Mails',
                'db_table': 'aa_bb_processed_mails',
            },
        ),
        migrations.RemoveField(
            model_name='bigbrotherconfig',
            name='mail_keywords',
        ),
        migrations.CreateModel(
            name='SusMailNote',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user_id', models.BigIntegerField(help_text='The AllianceAuth user ID who owns these characters')),
                ('note', models.TextField(help_text='The summary string of flags for this mail')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this note was created')),
                ('updated', models.DateTimeField(auto_now=True, help_text='When this note was last updated')),
                ('mail', models.OneToOneField(help_text='The mail this note refers to', on_delete=django.db.models.deletion.CASCADE, to='aa_bb.processedmail')),
            ],
            options={
                'verbose_name': 'Suspicious Mail Note',
                'verbose_name_plural': 'Suspicious Mail Notes',
                'db_table': 'aa_bb_sus_mail_notes',
            },
        ),
        migrations.CreateModel(
            name='SusContractNote',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user_id', models.BigIntegerField(help_text='The AllianceAuth user ID who owns these characters')),
                ('note', models.TextField(help_text='The summary string of flags for this contract')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When this note was created')),
                ('updated', models.DateTimeField(auto_now=True, help_text='When this note was last updated')),
                ('contract', models.OneToOneField(help_text='The contract this note refers to', on_delete=django.db.models.deletion.CASCADE, to='aa_bb.processedcontract')),
            ],
            options={
                'verbose_name': 'Suspicious Contract Note',
                'verbose_name_plural': 'Suspicious Contract Notes',
                'db_table': 'aa_bb_sus_contract_notes',
            },
        ),
    ]



# Generated by Django 4.2.21 on 2026-01-08 (Manual migration for consider_lowsec_hostile)

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0105_remove_tickettoolconfig_compliance_filter'),
    ]

    operations = [
        migrations.AddField(
            model_name='bigbrotherconfig',
            name='consider_lowsec_hostile',
            field=models.BooleanField(default=False, help_text='Consider all lowsec regions as hostile?', verbose_name='Consider Lowsec as Hostile'),
        ),
    ]



import html
import time
import json

from django.conf import settings
from django.contrib.auth.decorators import login_required, permission_required
from django.db import connection
from django.core.handlers.wsgi import WSGIRequest
from django.db.utils import OperationalError, ProgrammingError
from django.http import (
    JsonResponse,
    HttpResponseBadRequest,
    StreamingHttpResponse,
    HttpResponseForbidden,
)
from django.shortcuts import render, redirect, get_object_or_404
from django.views.decorators.http import require_POST
from django.core.cache import cache
from django.db.models import Q
from django_celery_beat.models import PeriodicTask
from django.utils import timezone

from celery import shared_task
from celery.exceptions import Ignore

from allianceauth.authentication.models import UserProfile, CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

from .forms import LeaveRequestForm
from .app_settings import (
    get_user_characters, get_entity_info, get_main_character_name,
    get_character_id, get_pings, aablacklist_active, send_status_embed,
    resolve_location_name, afat_active, discordbot_active, corptools_active,
    get_hostile_state
)
from .models import BigBrotherConfig, WarmProgress, LeaveRequest, ComplianceTicket, ComplianceTicketComment

from aa_bb.checks.awox import render_awox_kills_html
from aa_bb.checks.corp_changes import get_frequent_corp_changes
from aa_bb.checks.cyno import render_user_cyno_info_html
from aa_bb.checks.hostile_assets import render_assets
from aa_bb.checks.hostile_clones import render_clones
from aa_bb.checks.coalition_blacklist import get_external_blacklist_link
from aa_bb.checks.alliance_blacklist import get_alliance_blacklist_link
from aa_bb.checks.sus_contacts import render_contacts
from aa_bb.checks.sus_mails import (
    is_mail_row_hostile,
    get_cell_style_for_mail_cell,
    gather_user_mails,
    render_mails,
)
from aa_bb.checks.sus_trans import (
    get_user_transactions,
    is_transaction_hostile,
    gather_user_transactions,
    render_transactions,
    SUS_TYPES,
)
from .views_cb import CARD_DEFINITIONS

if aablacklist_active():
    from aa_bb.checks.add_to_blacklist import (
        get_add_to_blacklist_html,
        add_user_characters_to_blacklist,
        check_char_add_to_bl,
    )

from aa_bb.checks.sus_contracts import (
    get_user_contracts,
    is_contract_row_hostile,
    get_cell_style_for_contract_row,
    gather_user_contracts,
)
from aa_bb.checks.roles_and_tokens import render_user_roles_tokens_html
from aa_bb.checks.clone_state import render_character_states_html
from aa_bb.checks.skills import render_user_skills_html

logger = get_extension_logger(__name__)

try:
    if corptools_active():
        from corptools.models import Contract
    else:
        Contract = None
except ImportError:
    Contract = None


def get_allowed_alliance_id():
    cfg = BigBrotherConfig.get_solo()
    if not cfg.member_alliances:
        return None
    return int(cfg.member_alliances.split(",")[0].strip())


def get_allowed_coalition_alliance_ids():
    cfg = BigBrotherConfig.get_solo()
    if not cfg.whitelist_alliances:
        return set()

    return {
        int(a.strip())
        for a in cfg.whitelist_alliances.split(",")
        if a.strip().isdigit()
    }
try:
    ALLOWED_ALLIANCE_ID = get_allowed_alliance_id()
    ALLOWED_COALITION_ALLIANCE_IDS = get_allowed_coalition_alliance_ids()
except (OperationalError, ProgrammingError):
    ALLOWED_ALLIANCE_ID = None

CARD_DEFINITIONS = []

if aablacklist_active():
    CARD_DEFINITIONS.append(
        {"title": 'Add User to Blacklist', "key": "corp_bl"}
    )

CARD_DEFINITIONS += [
    {"title": 'Alliance Blacklist', "key": "alliance_bl"},
    {"title": 'Coalition Blacklist', "key": "external_bl"},
    {"title": 'Audit Compliance', "key": "compliance"},
    {"title": 'Player Corp History', "key": "freq_corp"},
    {"title": 'AWOX Kills', "key": "awox"},
    {"title": 'Omega State', "key": "clone_states"},
    {"title": 'Jump Clones', "key": "sus_clones"},
    {"title": 'Assets In Hostile Space', "key": "sus_asset"},
    {"title": 'Suspicious Contacts', "key": "sus_conta"},
    {"title": 'Suspicious Contracts', "key": "sus_contr"},
    {"title": 'Suspicious Mails', "key": "sus_mail"},
    {"title": 'Suspicious Transactions', "key": "sus_tra"},
    {"title": 'Cyno?', "key": "cyno"},
    {"title": 'Skills', "key": "skills"},
]



def get_available_cards():
    """Return card configurations filtered by settings and permissions."""
    cards = list(CARD_DEFINITIONS)
    try:
        cfg = BigBrotherConfig.get_solo()
    except (BigBrotherConfig.DoesNotExist, OperationalError, ProgrammingError):
        return cards

    if not corptools_active():
        corptools_cards = {
            "compliance", "freq_corp", "awox", "clone_states", "sus_clones",
            "sus_asset", "sus_conta", "sus_contr", "sus_mail", "sus_tra",
            "cyno", "skills"
        }
        cards = [card for card in cards if card["key"] not in corptools_cards]

    if not cfg.alliance_blacklist_url:
        cards = [card for card in cards if card["key"] != "alliance_bl"]

    if not cfg.external_blacklist_url:
        cards = [card for card in cards if card["key"] != "external_bl"]

    if not aablacklist_active():
        cards = [card for card in cards if card["key"] != "corp_bl"]

    return cards


def get_user_id(character_name):
    """Resolve an auth user ID from a character name, respecting member restrictions."""
    if not character_name:
        return None
    character_name = str(character_name)
    try:
        ownership = CharacterOwnership.objects.select_related('user__profile__main_character') \
            .get(character__character_name=character_name)

        cfg = BigBrotherConfig.get_solo()
        member_corps = {int(x) for x in (cfg.member_corporations or "").split(",") if x.strip().isdigit()}
        member_allis = {int(x) for x in (cfg.member_alliances or "").split(",") if x.strip().isdigit()}
        if member_corps or member_allis:
            main_char = getattr(ownership.user.profile, 'main_character', None)
            if not main_char or (main_char.corporation_id not in member_corps and main_char.alliance_id not in member_allis):
                return None

        return ownership.user.id
    except CharacterOwnership.DoesNotExist:
        return None

# Single-card loader
@login_required
@permission_required("aa_bb.basic_access")
def load_card(request):
    """Return the rendered HTML for a single dashboard card."""
    option = request.GET.get("option")
    idx    = request.GET.get("index")
    cards = get_available_cards()

    if option is None or idx is None:  # Card fetches require both parameters.
        return HttpResponseBadRequest("Missing parameters")

    try:
        idx      = int(idx)
        card_def = cards[idx]
    except (ValueError, IndexError):
        return HttpResponseBadRequest("Invalid card index")

    key   = card_def["key"]
    title = card_def["title"]
    logger.info(key)
    if key in ("sus_contr", "sus_mail","sus_tra"):  # Paginated cards handled separately via SSE/ajax.
        # handled via paginated endpoints
        return JsonResponse({"key": key, "title": title})

    target_user_id = get_user_id(option)
    if target_user_id is None:  # Unknown character selection.
        return JsonResponse({"error": "Unknown account"}, status=404)

    content, status = get_card_data(request, target_user_id, key)
    return JsonResponse({
        "title":   title,
        "content": content,
        "status":  status,
    })


# Bulk loader (fallback)
@login_required
@permission_required("aa_bb.basic_access")
def load_cards(request: WSGIRequest) -> JsonResponse:
    """Bulk-load every card for a selected user (fallback for legacy UI)."""
    selected_option = request.GET.get("option")
    user_id = get_user_id(selected_option)
    warm_entity_cache_task.delay(user_id)
    cards = []
    for card in get_available_cards():
        content, status = get_card_data(request, user_id, card["key"])
        cards.append({
            "title":   card["title"],
            "content": content,
            "status":  status,
        })
    return JsonResponse({"cards": cards})

@shared_task(bind=True)
def warm_entity_cache_task(self, user_id):
    """
    Gather mails, contracts, transactions; warm entity cache.
    Track progress in the DB via WarmProgress.
    """
    from .models import BigBrotherConfig
    cfg = BigBrotherConfig.get_solo()
    if not cfg.is_active or not cfg.is_warmer_active:
        return
    user_main = get_main_character_name(user_id) or str(user_id)
    qs = WarmProgress.objects.all()
    users = [
        {"user": wp.user_main, "current": wp.current, "total": wp.total}
        for wp in qs
    ]
    # Check for existing progress entry
    try:
        progress = WarmProgress.objects.get(user_main=user_main)
    except WarmProgress.DoesNotExist:
        progress = None

    # Determine if an existing warm job is currently making progress.
    if progress and progress.total > 0:
        first_current = progress.current
        logger.info(f"[{user_main}] detected in-progress run (current={first_current}); probing…")
        time.sleep(20)

        # Re-fetch progress record to see if current count has increased.
        try:
            progress = WarmProgress.objects.get(user_main=user_main)
            second_current = progress.current
        except WarmProgress.DoesNotExist:
            second_current = None

        # Abort if progress was detected; otherwise continue with the new task.
        if second_current != first_current:
            logger.info(
                f"[{user_main}] progress advanced from {first_current} to {second_current}; aborting new task."
            )
            raise Ignore(f"Task for {user_main} is already running.")
        else:
            logger.info(
                f"[{user_main}] no progress in 20 s (still {first_current}); continuing with new task."
            )

    # Build list of (entity_id, timestamp)
    entries = []
    contracts = gather_user_contracts(user_id)
    trans = gather_user_transactions(user_id)
    mails = gather_user_mails(user_id)
    candidates = []
    for c in contracts:
        issuer_id = get_character_id(c.issuer_name)
        candidates.append((issuer_id, getattr(c, "date_issued")))
        assignee = c.assignee_id or c.acceptor_id
        candidates.append((assignee, getattr(c, "date_issued")))
    for m in mails:
        candidates.append((m.from_id, getattr(m, "timestamp")))
        for mr in m.recipients.all():
            candidates.append((mr.recipient_id, getattr(m, "timestamp")))
    for t in trans:
        candidates.append((t.first_party_id, getattr(t, "date")))
        candidates.append((t.second_party_id, getattr(t, "date")))
    from django.db.models import Q
    from .models import EntityInfoCache
    query_filter = Q()
    for entity_id, as_of in candidates:
        query_filter |= Q(entity_id=entity_id, as_of=as_of)

    existing = set(
        EntityInfoCache.objects.filter(query_filter)
        .values_list('entity_id', 'as_of')
    )

    for candidate in candidates:
        if candidate not in existing:  # Only fetch entity info when cache lacks the tuple.
            entries.append(candidate)

    total = len(entries)
    logger.info(f"Starting warm cache for {user_main} ({total} entries)")

    # Initialize or update the progress record
    WarmProgress.objects.update_or_create(
        user_main=user_main,
        defaults={"current": 0, "total": total}
    )

    # Process each entry, updating the DB record
    for idx, (eid, ts) in enumerate(entries, start=1):
        WarmProgress.objects.filter(user_main=user_main).update(current=idx)
        get_entity_info(eid, ts)

    # Clean up when done
    WarmProgress.objects.filter(user_main=user_main).delete()
    logger.info(f"Completed warm cache for {user_main}")
    return total

@login_required
@permission_required("aa_bb.basic_access")
def warm_cache(request):
    """
    Endpoint to kick off warming for a given character name (option).
    Immediately registers a WarmProgress row so queued tasks also appear.
    """
    if not BigBrotherConfig.get_solo().is_warmer_active:  # Allow admins to disable the warmer
        return JsonResponse({"error": "Warmer disabled"}, status=403)
    option  = request.GET.get("option", "")
    user_id = get_user_id(option)
    if not user_id:  # Invalid character selection.
        return JsonResponse({"error": "Unknown account"}, status=400)

    # Pre-create progress record so queued jobs show up
    user_main = get_main_character_name(user_id) or str(user_id)
    WarmProgress.objects.get_or_create(
        user_main=user_main,
        defaults={"current": 0, "total": 0}
    )

    # Enqueue the celery task
    warm_entity_cache_task.delay(user_id)
    return JsonResponse({"started": True})


@login_required
@permission_required("aa_bb.basic_access")
def get_warm_progress(request):
    """
    AJAX endpoint returning all in-flight and queued warm-up info:
      {
        in_progress: bool,
        users: [ { user, current, total }, … ],
        queued: { count, names: [...] }
      }
    """
    qs = WarmProgress.objects.all()
    users = [
        {"user": wp.user_main, "current": wp.current, "total": wp.total}
        for wp in qs
    ]
    # Those still at current == 0 are queued/not yet started
    queued_names = [wp.user_main for wp in qs if wp.current == 0]

    #logger.debug(f"get_warm_progress → users={users}, queued={queued_names}")
    return JsonResponse({
        "in_progress": bool(users),
        "users": users,
        "queued": {
            "count": len(queued_names),
            "names": queued_names,
        },
    })

# Index view
@login_required
@permission_required("aa_bb.basic_access")
def index(request: WSGIRequest):
    """Render the dashboard shell plus dropdown options for authorized recruiters."""
    dropdown_options = []
    from .tasks_utils import format_task_name
    task_name = format_task_name('BB run regular updates')
    task = PeriodicTask.objects.filter(name=task_name).first()
    cfg = BigBrotherConfig.get_solo()
    cfg.is_active = True
    if not cfg.is_active:  # Guard against misconfigured BB.
        msg = (
            "Big Brother is currently inactive; please fill settings and enable the task"
        )
        return render(request, "aa_bb/disabled.html", {"message": msg})

    if request.user.has_perm("aa_bb.full_access"):  # Full-access sees every main character.
        qs = UserProfile.objects.exclude(main_character=None)
    elif request.user.has_perm("aa_bb.recruiter_access"):  # Recruiters see only guest states.
        guest_states = cfg.bb_guest_states.all()
        qs = UserProfile.objects.filter(state__in=guest_states).exclude(main_character=None)
    else:
        qs = None

    if qs is not None:  # Build dropdown choices only when the viewer has visibility.
        member_corps = {int(x) for x in (cfg.member_corporations or "").split(",") if x.strip().isdigit()}
        member_allis = {int(x) for x in (cfg.member_alliances or "").split(",") if x.strip().isdigit()}
        if member_corps or member_allis:
            qs = qs.filter(Q(main_character__corporation_id__in=member_corps) | Q(main_character__alliance_id__in=member_allis))

        dropdown_options = (
            qs.values_list("main_character__character_name", flat=True)
              .order_by("main_character__character_name")
        )

    context = {
        "dropdown_options": dropdown_options,
        "CARD_DEFINITIONS": get_available_cards(),
    }
    return render(request, "aa_bb/index.html", context)




# Paginated endpoints for Suspicious Contracts
@login_required
@permission_required("aa_bb.basic_access")
def list_contract_ids(request):
    """
    Return JSON list of all contract IDs and issue dates for the selected user.
    """
    option = request.GET.get("option")
    user_id = get_user_id(option)
    if user_id is None:  # Target selection must map to a known auth user.
        return JsonResponse({"error": "Unknown account"}, status=404)

    user_chars = get_user_characters(user_id)
    if Contract is not None:
        qs = Contract.objects.filter(
            character__character__character_id__in=user_chars
        ).order_by('-date_issued').values_list('contract_id', 'date_issued')
    else:
        qs = []

    contracts = [
        {'id': cid, 'date': dt.isoformat()} for cid, dt in qs
    ]
    return JsonResponse({'contracts': contracts})


@login_required
@permission_required("aa_bb.basic_access")
def check_contract_batch(request):
    """
    Check a slice of contracts for hostility by start/limit parameters.
    Returns JSON with `checked` count and list of `hostile_found`,
    each entry including a `cell_styles` dict for inline styling.
    Now uses gather_user_contracts + get_user_contracts(qs) on the full set.
    """
    option = request.GET.get("option")
    start  = int(request.GET.get("start", 0))
    limit  = int(request.GET.get("limit", 10))
    user_id = get_user_id(option)
    if user_id is None:  # Unknown selection -> 404.
        return JsonResponse({"error": "Unknown account"}, status=404)

    # 1) Ensure the full QuerySet is available
    cache_key = f"contract_qs_{user_id}"
    qs_all = cache.get(cache_key)
    if qs_all is None:  # Cache miss, gather the entire contract queryset now.
        qs_all = gather_user_contracts(user_id)
        cache.set(cache_key, qs_all, 300)

    # 2) Slice out just this batch of model instances
    batch_qs = qs_all[start:start + limit]

    # 3) Hydrate only this batch
    batch_map = get_user_contracts(batch_qs)

    HIDDEN = {
        'assignee_alliance_id', 'assignee_corporation_id',
        'issuer_alliance_id', 'issuer_corporation_id',
        'assignee_id', 'issuer_id', 'contract_id'
    }

    hostile = []
    for cid, row in batch_map.items():
        if is_contract_row_hostile(row):  # Only return rows flagged as hostile.
            # build style map for visible columns
            style_map = {
                col: get_cell_style_for_contract_row(col, row)
                for col in row
                if col not in HIDDEN
            }
            # package only the visible fields + styles
            payload = {col: row[col] for col in row if col not in HIDDEN}
            payload['cell_styles'] = style_map
            hostile.append(payload)

    return JsonResponse({
        'checked': len(batch_qs),
        'hostile_found': hostile
    })




@login_required
@permission_required("aa_bb.basic_access")
def stream_contracts_sse(request: WSGIRequest):
    """Push suspicious contract rows to the browser using server-sent events."""
    option = request.GET.get("option", "")
    user_id = get_user_id(option)
    if not user_id:  # SSE requires a valid user context.
        return HttpResponseBadRequest("Unknown account")

    if not corptools_active():
        return HttpResponseForbidden("Corptools required")

    try:
        qs    = gather_user_contracts(user_id)
        total = qs.count() if hasattr(qs, 'count') else len(qs)
        connection.close()
    except Exception as e:
        logger.error(f"Error initializing contract stream for {option}: {e}", exc_info=True)
        return HttpResponseBadRequest(f"Error loading contracts: {str(e)}")

    def generator():
        # Initial SSE heartbeat
        yield ": ok\n\n"
        processed = hostile_count = 0

        if total == 0:  # Nothing to scan, emit done immediately.
            # Notify client that processing completed with zero hostile hits
            yield "event: done\ndata:0\n\n"
            return

        for c in qs:
            processed += 1
            # Ping to keep connection alive
            yield ": ping\n\n"

            issued = getattr(c, "date_issued", timezone.now())
            issuer_id = c.issuer_name.eve_id
            yield ": ping\n\n"
            cid = c.contract_id
            if c.assignee_id != 0:  # Contracts may target assignee or acceptor; prefer assignee when set.
                assignee_id = c.assignee_id
            else:
                assignee_id = c.acceptor_id
            yield ": ping\n\n"
            #logger.info(f"getting info for {issuer_id}")
            iinfo     = get_entity_info(issuer_id, issued)
            yield ": ping\n\n"
            #logger.info(f"getting info for {assignee_id}")
            ainfo     = get_entity_info(assignee_id, issued)
            yield ": ping\n\n"

            # Hydrate just this one

            row = {
                'contract_id':              cid,
                'issued_date':              issued,
                'end_date':                 c.date_completed or c.date_expired,
                'contract_type':            c.contract_type,
                'issuer_name':              iinfo["name"],
                'issuer_id':                issuer_id,
                'issuer_corporation':       iinfo["corp_name"],
                'issuer_corporation_id':    iinfo["corp_id"],
                'issuer_alliance':          iinfo["alli_name"],
                'issuer_alliance_id':       iinfo["alli_id"],
                'assignee_name':            ainfo["name"],
                'assignee_id':              assignee_id,
                'assignee_corporation':     ainfo["corp_name"],
                'assignee_corporation_id':  ainfo["corp_id"],
                'assignee_alliance':        ainfo["alli_name"],
                'assignee_alliance_id':     ainfo["alli_id"],
                'status':                   c.status,
                'start_location':           resolve_location_name(getattr(c, "start_location_id", None)),
                'end_location':             resolve_location_name(getattr(c, "end_location_id", None)),
            }

            style_map = {
                col: get_cell_style_for_contract_row(col, row)
                for col in row
            }
            yield ": ping\n\n"
            row['cell_styles'] = style_map

            if is_contract_row_hostile(row):  # Emit rows that match hostile heuristics.
                hostile_count += 1
                tr_html = _render_contract_row_html(row)
                yield f"event: contract\ndata:{json.dumps(tr_html)}\n\n"

            # Progress update
            yield (
                "event: progress\n"
                f"data:{processed},{total},{hostile_count}\n\n"
            )
            connection.close()

        # Done
        yield "event: done\ndata:bye\n\n"

    resp = StreamingHttpResponse(generator(), content_type='text/event-stream')
    resp["Cache-Control"]     = "no-cache"
    resp["X-Accel-Buffering"] = "no"
    return resp



VISIBLE = [
    "sent_date", "subject",
    "sender_name", "sender_corporation", "sender_alliance",
    "recipient_names", "recipient_corps", "recipient_alliances",
    "content", "status",
]

VISIBLE_CONTR = [
    "issued_date", "end_date",
    "contract_type", "issuer_name", "issuer_corporation",
    "issuer_alliance", "assignee_name", "assignee_corporation",
    "assignee_alliance", "status", "start_location", "end_location",
]

def _render_contract_row_html(row: dict) -> str:
    """
    Render one hostile contract row, applying inline styles
    from row['cell_styles'] *or* from any hidden-ID–based flags.
    """
    cells = []

    # for any visible header like "issuer_name", map its ID column:
    def id_for(col):
        if col.endswith("_name"):  # Visible name columns pair with *_id fields.
            return col[:-5] + "_id"
        elif col.endswith("_corporation"):  # Map corp label to corp_id.
            return col[:-12] + "_corporation_id"
        elif col.endswith("_alliance"):  # Map alliance label to alliance_id.
            return col[:-9] + "_alliance_id"
        return None

    style_map = row.get('cell_styles', {})

    for col in VISIBLE_CONTR:
        val   = row.get(col, "")
        text  = html.escape(str(val))

        # first, try the direct style:
        style = style_map.get(col, "") or ""

        # render the cell
        if style:  # Inline styles highlight hostile issuers/assignees.
            cells.append(f'<td style="{style}">{text}</td>')
        else:
            cells.append(f'<td>{text}</td>')

    return "<tr>" + "".join(cells) + "</tr>"

def _render_mail_row_html(row: dict) -> str:
    """
    Render a single hostile mail row as <tr>…</tr> using only VISIBLE columns,
    applying red styling to any name whose ID is hostile.
    """
    cells = []
    cfg = BigBrotherConfig.get_solo()

    for col in VISIBLE:
        val = row.get(col, "")
        # recipients come as lists
        if isinstance(val, list):  # Expand recipient arrays to comma-separated spans.
            spans = []
            for i, item in enumerate(val):
                style = ""
                if col == "recipient_names":  # Hostile recipients get red styling.
                    rid = row["recipient_ids"][i]
                    if aablacklist_active():
                        if check_char_add_to_bl(rid):
                            style = "color:red;"
                elif col == "recipient_corps":  # Hostile corps -> red label.
                    cid = row["recipient_corp_ids"][i]
                    if cid and str(cid) in cfg.hostile_corporations:
                        style = "color:red;"
                elif col == "recipient_alliances":  # Hostile alliances -> red label.
                    aid = row["recipient_alliance_ids"][i]
                    if aid and str(aid) in cfg.hostile_alliances:
                        style = "color:red;"
                span = (
                    f'<span style="{style}">{html.escape(str(item))}</span>'
                    if style else
                    f'<span>{html.escape(str(item))}</span>'
                )
                spans.append(span)
            cell_html = ", ".join(spans)
        else:
            # single-valued columns: subject, content, sender_*
            style = ""
            if col.startswith("sender_"):  # Sender cells use existing cell style helper.
                style = get_cell_style_for_mail_cell(col, row, None)
            if col == "sender_name":
                for key in ["GM ","CCP "]:
                    if key in str(row["sender_name"]):  # Highlight official senders (GM/CCP) in red to stand out.
                        style = "color:red;"
            if style:  # Apply span styling when a highlight was requested.
                cell_html = f'<span style="{style}">{html.escape(str(val))}</span>'
            else:
                cell_html = html.escape(str(val))
        cells.append(f"<td>{cell_html}</td>")

    return "<tr>" + "".join(cells) + "</tr>"

@login_required
@permission_required("aa_bb.basic_access")
def stream_mails_sse(request):
    """Stream hostile mails one row at a time via SSE, hydrating sender+recipients."""
    option  = request.GET.get("option", "")
    user_id = get_user_id(option)
    if not user_id:  # Clients must specify a valid account to inspect.
        return HttpResponseBadRequest("Unknown account")

    if not corptools_active():
        return HttpResponseForbidden("Corptools required")

    try:
        qs    = gather_user_mails(user_id)
        total = qs.count()
        connection.close()
    except Exception as e:
        logger.error(f"Error initializing mail stream for {option}: {e}", exc_info=True)
        return HttpResponseBadRequest(f"Error loading mails: {str(e)}")

    def generator():
        # initial SSE heartbeat
        yield ": ok\n\n"
        processed = hostile_count = 0

        if total == 0:  # Nothing to stream -> immediately finish.
            # Notify client that streaming finished without hostile mails
            yield "event: done\ndata:0\n\n"
            return

        for m in qs:
            processed += 1
            # per-mail ping
            yield ": ping\n\n"

            sent = getattr(m, "timestamp", timezone.now())

            # 1) hydrate sender
            sender_id = m.from_id
            #logger.info(f"getting info for {sender_id}")
            sinfo     = get_entity_info(sender_id, sent)
            yield ": ping\n\n"  # immediately after expensive call

            # 2) hydrate each recipient
            recipient_ids           = []
            recipient_names         = []
            recipient_corps         = []
            recipient_corp_ids      = []
            recipient_alliances     = []
            recipient_alliance_ids  = []
            for mr in m.recipients.all():
                rid   = mr.recipient_id
                #logger.info(f"getting info for {rid}")
                rinfo = get_entity_info(rid, sent)
                yield ": ping\n\n"  # after each recipient lookup

                recipient_ids.append(rid)
                recipient_names.append(rinfo["name"])
                recipient_corps.append(rinfo["corp_name"])
                recipient_corp_ids.append(rinfo["corp_id"])
                recipient_alliances.append(rinfo["alli_name"])
                recipient_alliance_ids.append(rinfo["alli_id"])

            # build the single-mail row dict
            row = {
                "message_id":              m.id_key,
                "sent_date":               sent,
                "subject":                 m.subject or "",
                "sender_name":             sinfo["name"],
                "sender_id":               sender_id,
                "sender_corporation":      sinfo["corp_name"],
                "sender_corporation_id":   sinfo["corp_id"],
                "sender_alliance":         sinfo["alli_name"],
                "sender_alliance_id":      sinfo["alli_id"],
                "recipient_names":         recipient_names,
                "recipient_ids":           recipient_ids,
                "recipient_corps":         recipient_corps,
                "recipient_corp_ids":      recipient_corp_ids,
                "recipient_alliances":     recipient_alliances,
                "recipient_alliance_ids":  recipient_alliance_ids,
                "status":                  "Read" if m.is_read else "Unread",
            }

            # 3) check hostility and, if hostile, stream the <tr>
            if is_mail_row_hostile(row):  # Emit only hostile mail rows.
                hostile_count += 1
                tr = _render_mail_row_html(row)
                yield f"event: mail\ndata:{json.dumps(tr)}\n\n"

            # 4) final per-mail progress
            yield (
                "event: progress\n"
                f"data:{processed},{total},{hostile_count}\n\n"
            )
            connection.close()

        # done
        yield "event: done\ndata:bye\n\n"

    resp = StreamingHttpResponse(generator(),
                                 content_type="text/event-stream")
    resp["Cache-Control"]     = "no-cache"
    resp["X-Accel-Buffering"] = "no"
    return resp


@login_required
@permission_required("aa_bb.basic_access")
def stream_transactions_sse(request):
    """
    Stream hostile wallet‐transactions one <tr> at a time via SSE,
    hydrating first‐ and second‐party info on the fly.
    """
    option  = request.GET.get("option", "")
    user_id = get_user_id(option)
    if not user_id:  # Reject SSE connection when the pilot is unknown.
        return HttpResponseBadRequest("Unknown account")

    if not corptools_active():
        return HttpResponseForbidden("Corptools required")

    try:
        qs = gather_user_transactions(user_id)
        if hasattr(qs, 'order_by'):
            # Journal entries are best viewed in descending date order
            qs = qs.order_by('-date')
        total = qs.count() if hasattr(qs, 'count') else len(qs)
        logger.info(f"Transaction stream for {option}: found {total} total transactions")
        connection.close()
    except Exception as e:
        logger.error(f"Error initializing transaction stream for {option}: {e}", exc_info=True)
        return HttpResponseBadRequest(f"Error loading transactions: {str(e)}")

    # Hidden columns for the transactions table
    HIDDEN        = {
        'first_party_id','second_party_id',
        'first_party_corporation_id','second_party_corporation_id',
        'first_party_alliance_id','second_party_alliance_id',
        'entry_id'
    }

    def generator():
        yield ": ok\n\n"                # initial heartbeat
        processed = hostile_count = 0

        if total == 0:  # No transactions -> stop immediately.
            # Notify client that processing ended without hostile entries
            yield "event: done\ndata:0\n\n"
            return

        # Determine headers from a single hydrated row (after empty check)
        sample_map = get_user_transactions(qs[:1])
        sample_row = next(iter(sample_map.values()), None)
        if sample_row:  # Derive headers from real data when available.
            headers = [h for h in sample_row.keys() if h not in HIDDEN]
        else:
            # Fallback to a safe default when sampling finds nothing
            headers = [
                'date', 'amount', 'balance', 'description', 'reason',
                'first_party_name', 'first_party_corporation', 'first_party_alliance',
                'second_party_name', 'second_party_corporation', 'second_party_alliance',
                'context', 'type',
            ]

        # Emit table header row once
        header_html = (
            "<tr>" +
            "".join(f"<th>{html.escape(h.replace('_',' ').title())}</th>" for h in headers) +
            "</tr>"
        )
        yield f"event: header\ndata:{json.dumps(header_html)}\n\n"

        cfg = BigBrotherConfig.get_solo()

        user_chars = get_user_characters(user_id)
        user_ids = set(user_chars.keys())

        batch_size = 100
        for i in range(0, total, batch_size):
            batch = qs[i:i + batch_size]
            rows_map = get_user_transactions(batch)
            # Sort keys to maintain some order if possible, though date desc is better
            sorted_keys = sorted(rows_map.keys(), key=lambda k: rows_map[k]['date'], reverse=True)

            for eid in sorted_keys:
                row = rows_map[eid]
                processed += 1
                if processed % 10 == 0:
                    yield ": ping\n\n"         # keep‐alive

                is_hostile = is_transaction_hostile(row, user_ids)
                if processed <= 5:  # Log first 5 for debugging
                    logger.info(f"TX {eid}: type={row.get('type')}, fp={row.get('first_party_id')}, sp={row.get('second_party_id')}, hostile={is_hostile}")

                if is_hostile:  # Only push rows that meet hostility rules.
                    hostile_count += 1

                    # build the <tr> using same style logic as render_transactions()
                    cells = []
                    for col in headers:
                        val = row.get(col, "")
                        text = html.escape(str(val))
                        style = ""
                        # type‐based red
                        if col == 'type':
                            if any(st in row['type'] for st in SUS_TYPES):
                                style = 'color:red;'
                            if cfg.show_market_transactions:
                                if "market_escrow" in row['type'] or "market_transaction" in row['type']:
                                    style = 'color:red;'
                        # first/second party name
                        if col in ('first_party_name','second_party_name'):
                            pid = row[col.replace("_name", "_id")]
                            if get_hostile_state(pid, 'character', when=row['date']):
                                style = 'color:red;'
                        # corps & alliances
                        if col.endswith('corporation'):
                            cid = row[f"{col}_id"]
                            if cid and get_hostile_state(cid, 'corporation', when=row['date']):
                                style = 'color:red;'
                        if col.endswith('alliance'):
                            aid = row[f"{col}_id"]
                            if aid and get_hostile_state(aid, 'alliance', when=row['date']):
                                style = 'color:red;'
                        def make_td(text, style=""):
                            style_attr = f' style="{style}"' if style else ""
                            return f"<td{style_attr}>{text}</td>"
                        cells.append(make_td(text, style))
                    tr_html = "<tr>" + "".join(cells) + "</tr>"
                    yield f"event: transaction\ndata:{json.dumps(tr_html)}\n\n"

                # progress update every few rows to avoid flood
                if processed % 10 == 0 or processed == total:
                    yield (
                        "event: progress\n"
                        f"data:{processed},{total},{hostile_count}\n\n"
                    )
            connection.close()

        # Done
        yield "event: done\ndata:bye\n\n"

    resp = StreamingHttpResponse(generator(), content_type='text/event-stream')
    resp["Cache-Control"]     = "no-cache"
    resp["X-Accel-Buffering"] = "no"
    return resp


# Card data helper

def get_card_data(request, target_user_id: int, key: str):
    """Return card HTML and status tuple for the specified key."""

    if key == "compliance":  # Role/token compliance overview.
        content = render_user_roles_tokens_html(target_user_id)
        status = not (content and "danger" in content)

    elif key == "corp_bl":  # Inline corp blacklist check (with add links).
        issuer_id = request.user.id
        content   = get_add_to_blacklist_html(request, issuer_id, target_user_id) or "a"
        status    = not (content and "danger" in content)

    elif key == "alliance_bl":
        content = get_alliance_blacklist_link()
        status = True

    elif key == "external_bl":
        content = get_external_blacklist_link()
        status = True

    elif key == "freq_corp":  # Show frequent corporation changes timeline.
        content = get_frequent_corp_changes(target_user_id)
        status  = "danger" not in content

    elif key == "awox":  # Highlight kills where corp mates attacked each other.
        content = render_awox_kills_html(target_user_id)
        status  = content is None

    elif key == "clone_states":  # Clone state availability (alpha/omega).
        content = render_character_states_html(target_user_id)
        status = not (content and "danger" in content)

    elif key == "sus_clones":  # Flag clones located in hostile space.
        content = render_clones(target_user_id)
        status  = not (content and any(w in content for w in ("danger", "warning")))

    elif key == "sus_asset":  # Summarize assets currently stranded in hostile systems.
        content = render_assets(target_user_id)
        status  = not (content and "danger" in content)

    elif key == "sus_conta":  # Suspicious contact list card.
        content = render_contacts(target_user_id)
        status  = not (content and "danger" in content)

    elif key == "sus_mail":  # Suspicious mail preview card.
        content = render_mails(target_user_id)
        status  = not (content and "danger" in content)

    elif key == "sus_tra":  # Suspicious transaction summary card.
        content = render_transactions(target_user_id)
        status  = not content

    elif key == "cyno":  # Cyno readiness / history panel.
        content = render_user_cyno_info_html(target_user_id)
        status  = not (content and "danger" in content)

    elif key == "skills":  # Training gaps summary.
        content = render_user_skills_html(target_user_id)
        status  = not (content and "danger" in content)

    else:
        content = "WiP"
        status  = True

    return content, status


@require_POST
@permission_required("can_blacklist_characters")
def add_blacklist_view(request):
    """POST endpoint to add all of a target's characters to the corp blacklist."""
    issuer_id = int(request.POST["issuer_user_id"])
    target_id = int(request.POST["target_user_id"])
    reason    = request.POST.get("reason", "")
    added = add_user_characters_to_blacklist(
        issuer_user_id=issuer_id,
        target_user_id=target_id,
        reason=reason
    )
    return redirect(
        request.META.get("HTTP_REFERER", "/"),
        message=f"Blacklisted: {', '.join(added)}"
    )


@login_required
@permission_required("aa_bb.can_access_loa")
def loa_loa(request):
    """Display the LoA dashboard for the requesting user."""
    cfg = BigBrotherConfig.get_solo()
    if not cfg.is_loa_active:  # Feature toggle to hide LoA entirely.
        return render(request, "loa/disabled.html")
    user_requests = LeaveRequest.objects.filter(user=request.user).order_by('-created_at')
    return render(request, "loa/index.html", {"loa_requests": user_requests})

@login_required
@permission_required("aa_bb.can_view_all_loa")
def loa_admin(request):
    """Administrative LoA queue view with filtering."""
    cfg = BigBrotherConfig.get_solo()
    if not cfg.is_loa_active:  # Hide admin view when LoA disabled globally.
        return render(request, "loa/disabled.html")
    # Filtering
    qs = LeaveRequest.objects.select_related('user').order_by('-created_at')

    member_corps = {int(x) for x in (cfg.member_corporations or "").split(",") if x.strip().isdigit()}
    member_allis = {int(x) for x in (cfg.member_alliances or "").split(",") if x.strip().isdigit()}
    if member_corps or member_allis:
        qs = qs.filter(Q(user__profile__main_character__corporation_id__in=member_corps) | Q(user__profile__main_character__alliance_id__in=member_allis))

    user_filter   = request.GET.get('user')
    status_filter = request.GET.get('status')

    if user_filter:  # Narrow to a single user's requests.
        qs = qs.filter(user__id=user_filter)
    if status_filter:  # Filter by request status (pending/approved/etc).
        qs = qs.filter(status=status_filter)

    # Build dropdown options from existing requests
    users_in_requests_qs = LeaveRequest.objects.all()
    member_corps = {int(x) for x in (cfg.member_corporations or "").split(",") if x.strip().isdigit()}
    member_allis = {int(x) for x in (cfg.member_alliances or "").split(",") if x.strip().isdigit()}
    if member_corps or member_allis:
        users_in_requests_qs = users_in_requests_qs.filter(Q(user__profile__main_character__corporation_id__in=member_corps) | Q(user__profile__main_character__alliance_id__in=member_allis))

    users_in_requests = (
        users_in_requests_qs.values_list('user__id', 'user__username')
                            .distinct()
    )

    context = {
        'loa_requests': qs,
        'users': users_in_requests,
        'status_choices': LeaveRequest.STATUS_CHOICES,
        'current_user': user_filter,
        'current_status': status_filter,
    }
    return render(request, "loa/admin.html", context)

@login_required
@permission_required("aa_bb.can_access_loa")
def loa_request(request):
    """Handle LoA request creation form (GET/POST)."""
    cfg = BigBrotherConfig.get_solo()
    if not cfg.is_loa_active:  # Respect feature toggle.
        return render(request, "loa/disabled.html")

    if request.method == 'POST':  # Form submission branch.
        form = LeaveRequestForm(request.POST)
        if form.is_valid():  # Save request and ping staff.
            main_char = get_main_character_name(request.user.id)
            # 2) save with main_character
            lr = form.save(commit=False)
            lr.user = request.user
            lr.main_character = main_char
            lr.save()

            # 3) send webhook with character
            hook = cfg.loawebhook
            send_status_embed(
                subject="LoA Request",
                lines=[
                    f"{get_pings('LoA Request')} {main_char} requested LOA:",
                    f"- from **{lr.start_date}**",
                    f"- to **{lr.end_date}**",
                    f"- reason: **{lr.reason}**"
                ],
                color=0x3498db,
                hook=hook
            )

            return redirect('loa:index')
        else:
            form.add_error(None, "Please fill in all fields correctly.")
    else:
        form = LeaveRequestForm()

    return render(request, 'loa/request.html', {'form': form})

@login_required
@permission_required("aa_bb.can_access_loa")
def delete_request(request, pk):
    """Allow a user to delete their own pending LoA."""
    if request.method == 'POST':  # Only accept POST to mutate state.
        lr = get_object_or_404(LeaveRequest, pk=pk, user=request.user)
        if lr.user != request.user:  # Safety net in case of tampering.
            return HttpResponseForbidden("You may only delete your own requests.")
        elif lr.status == 'pending':  # Only pending requests may be removed.
            lr.delete()
            cfg = BigBrotherConfig.get_solo()
            hook = cfg.loawebhook
            send_status_embed(
                subject="LoA Deleted",
                lines=[
                    f"{get_pings('LoA Changed Status')} {lr.main_character} deleted their LOA:",
                    f"- from **{lr.start_date}**",
                    f"- to **{lr.end_date}**",
                    f"- reason: **{lr.reason}**"
                ],
                color=0x3498db,
                hook=hook
            )
    return redirect('loa:index')

@login_required
@permission_required("aa_bb.can_manage_loa")
def delete_request_admin(request, pk):
    """Admin-only delete path for any LoA request."""
    if request.method == 'POST':  # Guard mutation behind POST.
        cfg = BigBrotherConfig.get_solo()
        lr = get_object_or_404(LeaveRequest, pk=pk)
        member_corps = {int(x) for x in (cfg.member_corporations or "").split(",") if x.strip().isdigit()}
        member_allis = {int(x) for x in (cfg.member_alliances or "").split(",") if x.strip().isdigit()}
        if member_corps or member_allis:
            main_char = getattr(lr.user.profile, 'main_character', None)
            if not main_char or (main_char.corporation_id not in member_corps and main_char.alliance_id not in member_allis):
                return HttpResponseForbidden("Target user is not in a member corporation/alliance.")

        lr.delete()
        hook = cfg.loawebhook
        userrr = get_main_character_name(request.user.id)
        send_status_embed(
            subject="LoA Deleted by Admin",
            lines=[
                f"{get_pings('LoA Changed Status')} {userrr} deleted {lr.main_character}'s LOA:",
                f"- from **{lr.start_date}**",
                f"- to **{lr.end_date}**",
                f"- reason: **{lr.reason}**"
            ],
            color=0x3498db,
            hook=hook
        )
    return redirect('loa:admin')

@login_required
@permission_required("aa_bb.can_manage_loa")
def approve_request(request, pk):
    """Mark an LoA approved and notify Discord."""
    if request.method == 'POST':  # Only process POST actions.
        cfg = BigBrotherConfig.get_solo()
        lr = get_object_or_404(LeaveRequest, pk=pk)
        member_corps = {int(x) for x in (cfg.member_corporations or "").split(",") if x.strip().isdigit()}
        member_allis = {int(x) for x in (cfg.member_alliances or "").split(",") if x.strip().isdigit()}
        if member_corps or member_allis:
            main_char = getattr(lr.user.profile, 'main_character', None)
            if not main_char or (main_char.corporation_id not in member_corps and main_char.alliance_id not in member_allis):
                return HttpResponseForbidden("Target user is not in a member corporation/alliance.")

        lr.status = 'approved'
        lr.save()
        hook = cfg.loawebhook
        userrr = get_main_character_name(request.user.id)
        send_status_embed(
            subject="LoA Approved",
            lines=[
                f"{get_pings('LoA Changed Status')} {userrr} approved {lr.main_character}'s LOA:",
                f"- from **{lr.start_date}**",
                f"- to **{lr.end_date}**",
                f"- reason: **{lr.reason}**"
            ],
            color=0x3498db,
            hook=hook
        )
    return redirect('loa:admin')

@login_required
@permission_required("aa_bb.can_manage_loa")
def deny_request(request, pk):
    """Mark an LoA denied and notify Discord."""
    if request.method == 'POST':  # Only mutate via POST requests.
        cfg = BigBrotherConfig.get_solo()
        lr = get_object_or_404(LeaveRequest, pk=pk)
        member_corps = {int(x) for x in (cfg.member_corporations or "").split(",") if x.strip().isdigit()}
        member_allis = {int(x) for x in (cfg.member_alliances or "").split(",") if x.strip().isdigit()}
        if member_corps or member_allis:
            main_char = getattr(lr.user.profile, 'main_character', None)
            if not main_char or (main_char.corporation_id not in member_corps and main_char.alliance_id not in member_allis):
                return HttpResponseForbidden("Target user is not in a member corporation/alliance.")

        lr.status = 'denied'
        lr.save()
        hook = cfg.loawebhook
        userrr = get_main_character_name(request.user.id)
        send_status_embed(
            subject="LoA Denied",
            lines=[
                f"{get_pings('LoA Changed Status')} {userrr} denied {lr.main_character}'s LOA:",
                f"- from **{lr.start_date}**",
                f"- to **{lr.end_date}**",
                f"- reason: **{lr.reason}**"
            ],
            color=0x3498db,
            hook=hook
        )
    return redirect('loa:admin')


@login_required
@permission_required("aa_bb.ticket_manager")
def ticket_list(request):
    """List compliance tickets."""
    from django.db.models import Count

    tab = request.GET.get('tab', 'open')

    if tab == 'all':
        tickets = ComplianceTicket.objects.all()
    elif tab == 'exceptions':
        tickets = ComplianceTicket.objects.filter(is_exception=True)
    elif tab == 'resolved':
        tickets = ComplianceTicket.objects.filter(is_resolved=True)
    else:  # 'open' is default
        tickets = ComplianceTicket.objects.filter(is_resolved=False, is_exception=False)

    if not afat_active():
        tickets = tickets.exclude(reason="paps_check")

    if not discordbot_active():
        tickets = tickets.exclude(reason="discord_check")

    tickets = tickets.select_related('user__profile__main_character').annotate(comment_count=Count('comments'))
    return render(request, 'aa_bb/ticket_list.html', {'tickets': tickets, 'current_tab': tab})


@login_required
@permission_required("aa_bb.ticket_manager")
def ticket_view(request, pk):
    """View details and comments for a specific ticket."""
    ticket = get_object_or_404(
        ComplianceTicket.objects.select_related('user__profile__main_character'),
        pk=pk
    )
    if ticket.reason == "paps_check" and not afat_active():
        return HttpResponseForbidden("PAP compliance tickets are hidden as afat is not active.")
    comments = ticket.comments.all().select_related('user__profile__main_character')
    return render(request, 'aa_bb/ticket_view.html', {'ticket': ticket, 'comments': comments})


@login_required
@permission_required("aa_bb.ticket_manager")
@require_POST
def ticket_resolve(request, pk):
    """Resolve a ticket via the UI."""
    ticket = get_object_or_404(ComplianceTicket, pk=pk)
    from .tasks_tickets import close_ticket, close_char_removed_ticket

    if ticket.reason in ["char_removed", "awox_kill"]:
        close_char_removed_ticket(ticket, user=request.user)
    else:
        close_ticket(ticket, user=request.user)

    return redirect('aa_bb:ticket_view', pk=pk)


@login_required
@permission_required("aa_bb.ticket_manager")
@require_POST
def ticket_reopen(request, pk):
    """Reopen a resolved ticket."""
    ticket = get_object_or_404(ComplianceTicket, pk=pk)
    from .tasks_tickets import reopen_ticket
    reopen_ticket(ticket, user=request.user)

    return redirect('aa_bb:ticket_view', pk=pk)


@login_required
@permission_required("aa_bb.ticket_manager")
@require_POST
def ticket_mark_exception(request, pk):
    """Mark a ticket as an exception."""
    ticket = get_object_or_404(ComplianceTicket, pk=pk)
    reason = request.POST.get('reason', '').strip()

    from .tasks_tickets import mark_ticket_exception
    mark_ticket_exception(ticket, request.user, reason)

    return redirect('aa_bb:ticket_view', pk=pk)


@login_required
@permission_required("aa_bb.ticket_manager")
@require_POST
def ticket_clear_exception(request, pk):
    """Clear exception status from a ticket."""
    ticket = get_object_or_404(ComplianceTicket, pk=pk)

    from .tasks_tickets import clear_ticket_exception
    clear_ticket_exception(ticket, request.user)

    return redirect('aa_bb:ticket_view', pk=pk)


@login_required
@permission_required("aa_bb.ticket_manager")
@require_POST
def ticket_add_comment(request, pk):
    """Add a comment to a ticket and optionally post to Discord."""
    ticket = get_object_or_404(ComplianceTicket, pk=pk)
    comment_text = request.POST.get('comment')
    if comment_text:
        from .tasks_tickets import add_ticket_comment
        add_ticket_comment(ticket, request.user, comment_text)

    return redirect('aa_bb:ticket_view', pk=pk)



# Generated by Django 4.2.23 on 2026-01-02 23:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0098_bigbrotherconfig_exclude_high_sec_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='tickettoolconfig',
            name='hr_forum_webhook',
            field=models.URLField(blank=True, help_text='Discord webhook for creating forum threads for compliance issues (Fallback if discordbot is not installed)', null=True, verbose_name='HR Forum Webhook'),
        ),
    ]



# Generated by Django 4.2.23 on 2025-08-05 10:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0026_alter_general_options_alter_bigbrotherconfig_id_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False, 'permissions': (('basic_access', 'Can access this app'), ('full_access', 'Can view all main characters'), ('recruiter_access', 'Can view guest main characters only'), ('can_blacklist_characters', 'Can add characters to blacklist'), ('can_access_loa', 'Can submit a Leave Of Absence request'), ('can_view_all_loa', 'Can view all Leave Of Absence requests'), ('can_manage_loa', 'Can manage Leave Of Absence requests'))},
        ),
        migrations.AddField(
            model_name='bigbrotherconfig',
            name='is_loa_active',
            field=models.BooleanField(default=False, help_text='has the Leave of Absence module been activated/deactivated?'),
        ),
        migrations.AddField(
            model_name='bigbrotherconfig',
            name='loawebhook',
            field=models.URLField(blank=True, help_text='Discord webhook for sending Leave of Absence', null=True),
        ),
    ]



"""
Identify where members keep assets in space and flag hostile owners.

The routines below are used both for the HTML renderings and faux-alerts
that can be sent when a user has assets in systems owned by enemies.
"""

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger
from django.contrib.auth.models import User
from ..app_settings import (
    get_system_owner,
    is_nullsec,
    is_player_structure,
    get_safe_entities,
    resolve_location_name,
    resolve_location_system_id,
    is_highsec,
    is_lowsec,
    corptools_active,
    is_hostile_unified,
    is_ship,
)
from django.utils import timezone
from ..models import BigBrotherConfig
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from typing import List, Optional, Dict

logger = get_extension_logger(__name__)

try:
    if corptools_active():
        from corptools.models import CharacterAudit, CharacterAsset, EveLocation
    else:
        CharacterAudit = None
        CharacterAsset = None
        EveLocation = None
except ImportError:
    CharacterAudit = None
    CharacterAsset = None
    EveLocation = None


def _parse_id_list(value: Optional[str]) -> set[int]:
    if not value:
        return set()
    return {int(x) for x in value.split(",") if x.strip().isdigit()}


def get_asset_locations(user_id: int) -> Dict[int, dict]:
    """
    Return a dict mapping system IDs to a dict containing their name and a list of locations
    (stations/structures) where any of the given user's characters has one or more assets.
    """
    if not corptools_active() or CharacterAudit is None:
        return {}
    try:
        user = User.objects.get(pk=user_id)
    except User.DoesNotExist:
        return {}

    system_map: Dict[int, dict] = {}
    _loc_sys_cache = {}
    _loc_name_cache = {}

    def add_asset(system_obj, location_id, location_name, char_id, char_name, type_id, type_name):
        """Store the asset details organized by system and location."""
        key = None
        sys_name = None

        if system_obj:
            key = getattr(system_obj, "pk", None)
            sys_name = system_obj.name
        elif location_id:
            # Attempt to resolve system name
            if location_id in _loc_sys_cache:
                key = _loc_sys_cache[location_id]
            else:
                key = resolve_location_system_id(location_id)
                _loc_sys_cache[location_id] = key

            if key:
                if key in _loc_name_cache:
                    sys_name = _loc_name_cache[key]
                else:
                    sys_name = resolve_location_name(key)
                    _loc_name_cache[key] = sys_name

        if not key:
            return

        if key not in system_map:
            system_map[key] = {"name": sys_name, "locations": {}}

        # Determine location name to use as key/label
        loc_key = location_id or 0
        if loc_key not in system_map[key]["locations"]:
            if location_name:
                resolved_loc_name = location_name
            else:
                if loc_key in _loc_name_cache:
                    resolved_loc_name = _loc_name_cache[loc_key]
                else:
                    resolved_loc_name = resolve_location_name(loc_key) or f"Unknown Location {loc_key}"
                    _loc_name_cache[loc_key] = resolved_loc_name

            system_map[key]["locations"][loc_key] = {
                "name": resolved_loc_name,
                "assets": [],
            }

        system_map[key]["locations"][loc_key]["assets"].append({
            "char_id": char_id,
            "char_name": char_name,
            "type_id": type_id,
            "type_name": type_name,
        })

    # for each EVE character owned by this user
    for co in CharacterOwnership.objects.filter(user=user).select_related("character"):
        try:
            char_audit = CharacterAudit.objects.get(character=co.character)
        except CharacterAudit.DoesNotExist:
            continue

        assets = (
            CharacterAsset.objects.select_related(
                "location_name__system",
                "type_name__group__category",
            )
            .filter(character=char_audit)
            .exclude(location_flag__iexact="solar_system")
        )

        for asset in assets:
            if (asset.location_flag or "").lower() == "assetsafety":
                continue

            loc = asset.location_name
            system_obj = getattr(loc, "system", None) if loc else None

            # Optimization: use cached name from select_related if available
            loc_name = None
            if loc:
                loc_name = loc.location_name

            if not loc_name:
                loc_key = asset.location_id
                if loc_key in _loc_name_cache:
                    loc_name = _loc_name_cache[loc_key]
                else:
                    loc_name = resolve_location_name(loc_key)
                    _loc_name_cache[loc_key] = loc_name

            if not asset.type_name:
                continue

            add_asset(
                system_obj, asset.location_id, loc_name,
                co.character.character_id, co.character.character_name,
                asset.type_name.type_id, asset.type_name.name
            )

    return system_map


def get_hostile_asset_locations(user_id: int) -> Dict[str, dict]:
    """
    Returns a mapping of system display name -> structured hostile data
    for systems where the user's characters have assets in space and the
    system is considered hostile under the unified processor logic.
    """
    systems = get_asset_locations(user_id)
    if not systems:
        return {}

    hostile_map: Dict[str, dict] = {}
    from ..app_settings import get_safe_entities
    safe_entities = get_safe_entities()
    _hostile_memo = {}

    for system_id, data in systems.items():
        system_name = data.get("name")
        display_name = system_name or f"Unknown ({system_id})"

        # We need the system owner info for the summary
        owner_info = get_system_owner({"id": system_id, "name": display_name})
        oname = owner_info.get("owner_name", "Unresolvable") if owner_info else "Unresolvable"
        rname = owner_info.get("region_name", "Unknown Region") if owner_info else "Unknown Region"

        system_has_hostile = False
        # structured data: list of records
        records = []

        # Check each location in this system
        for loc_id, loc_data in data.get("locations", {}).items():
            loc_name = loc_data["name"]

            # Map of char_name -> set of ships for this specific location
            char_ships_at_loc = {}
            location_has_hostile = False

            for asset in loc_data.get("assets", []):
                # Memoize check results for character+location+type
                memo_key = (asset["char_id"], loc_id, system_id, asset["type_id"])
                if memo_key in _hostile_memo:
                    is_hostile = _hostile_memo[memo_key]
                else:
                    # Use unified check for each asset
                    is_hostile = is_hostile_unified(
                        involved_ids=[asset["char_id"]],
                        location_id=loc_id,
                        system_id=system_id,
                        is_asset=True,
                        asset_type_id=asset["type_id"],
                        when=timezone.now(),
                        safe_entities=safe_entities
                    )
                    _hostile_memo[memo_key] = is_hostile

                if is_hostile:
                    system_has_hostile = True
                    location_has_hostile = True
                    cname = asset["char_name"]
                    if is_ship(asset["type_id"]):
                        char_ships_at_loc.setdefault(cname, set()).add(asset["type_name"])
                    else:
                        # Ensure character is recorded even if no ships (e.g. just modules/items)
                        char_ships_at_loc.setdefault(cname, set())

            if location_has_hostile:
                for cname, ships in char_ships_at_loc.items():
                    records.append({
                        "char_name": cname,
                        "location_name": loc_name,
                        "ships": sorted(list(ships))
                    })

        if system_has_hostile:
            hostile_map[display_name] = {
                "owner": oname,
                "region": rname,
                "records": records
            }
            logger.info(f"Hostile asset system: {display_name} owned by {oname}")

    return hostile_map



def render_assets(user_id: int) -> Optional[str]:
    """
    Returns an HTML table listing each system where the user's characters have assets,
    the system's sovereign owner, and highlights in red any asset considered hostile.
    """
    try:
        systems = get_asset_locations(user_id)
        if not systems:
            return None

        rows: List[Dict] = []
        from ..app_settings import get_safe_entities
        safe_entities = get_safe_entities()
        _hostile_memo = {}

        for system_id, data in systems.items():
            system_name = data.get("name")
            display_name = system_name or f"Unknown ({system_id})"

            # Base system owner info for the table
            owner_info = get_system_owner({"id": system_id, "name": display_name})
            oname = owner_info.get("owner_name", "Unresolvable") if owner_info else "Unresolvable"
            region_name = owner_info.get("region_name", "Unknown Region") if owner_info else "Unknown Region"

            # Iterate locations inside system
            for loc_id, loc_data in data.get("locations", {}).items():
                loc_name = loc_data["name"]

                # Check each asset group (char/type combo)
                # Actually we can group by char for rendering
                char_assets = {}
                for asset in loc_data.get("assets", []):
                    char_name = asset["char_name"]
                    if char_name not in char_assets:
                        char_assets[char_name] = {"ships": [], "is_hostile": False, "char_id": asset["char_id"]}

                    # Memoize check results for character+location+type
                    memo_key = (asset["char_id"], loc_id, system_id, asset["type_id"])
                    if memo_key in _hostile_memo:
                        is_hostile = _hostile_memo[memo_key]
                    else:
                        # Check if this specific asset is hostile
                        is_hostile = is_hostile_unified(
                            involved_ids=[asset["char_id"]],
                            location_id=loc_id,
                            system_id=system_id,
                            is_asset=True,
                            asset_type_id=asset["type_id"],
                            when=timezone.now(),
                            safe_entities=safe_entities
                        )
                        _hostile_memo[memo_key] = is_hostile

                    if is_hostile:
                        char_assets[char_name]["is_hostile"] = True
                        if is_ship(asset["type_id"]):
                            char_assets[char_name]["ships"].append(asset["type_name"])

                for char_name, cdata in char_assets.items():
                    ship_str = ", ".join(sorted(cdata["ships"])) if cdata["ships"] else ""
                    rows.append({
                        "system": display_name,
                        "location": loc_name,
                        "character": char_name,
                        "owner": oname,
                        "region": region_name,
                        "hostile": cdata["is_hostile"],
                        "ships": ship_str,
                    })

        if not rows:
            return "<p>No hostile assets found.</p>"

        # Sort rows: hostile first, then by system, location, character
        rows.sort(key=lambda x: (not x["hostile"], x["system"], x["location"], x["character"]))

        html_output = '<table class="table table-striped table-hover stats">'
        html_output += (
            '<thead>'
            '  <tr>'
            '      <th style="width: 15%">System</th>'
            '      <th style="width: 20%">Station</th>'
            '      <th style="width: 15%">Character</th>'
            '      <th style="width: 15%">Owner</th>'
            '      <th style="width: 15%">Region</th>'
            '      <th style="width: 20%">Hostile Asset</th>'
            '  </tr>'
            '</thead>'
            '<tbody>'
        )

        for row in rows:
            system_cell = row["system"]
            owner_cell = row["owner"]
            region_cell = row["region"]
            hostile_ship = row["ships"] if row["hostile"] else ""

            if row["hostile"]:
                owner_cell = mark_safe(f'<span class="text-danger">{owner_cell}</span>')

            html_output += format_html(
                '   <tr>'
                '       <td>{}</td>'
                '       <td>{}</td>'
                '       <td>{}</td>'
                '       <td>{}</td>'
                '       <td>{}</td>'
                '       <td>{}</td>'
                '   </tr>',
                system_cell,
                row["location"],
                row["character"],
                owner_cell,
                region_cell,
                hostile_ship,
            )

        html_output += '</tbody></table>'
        return html_output
    except Exception as e:
        logger.exception(f"Error rendering assets for user {user_id}")
        return f"<p class='text-danger'>Error rendering assets: {str(e)}</p>"



# Generated by Django 4.2.23 on 2025-08-05 10:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0028_alter_bigbrotherconfig_is_loa_active'),
    ]

    operations = [
        migrations.AddField(
            model_name='leaverequest',
            name='main_character',
            field=models.CharField(blank=True, help_text="The user's primary character when they made the request", max_length=100),
        ),
    ]



# Generated by Django 4.2.23 on 2025-08-27 08:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0063_alter_complianceticket_user'),
    ]

    operations = [
        migrations.AlterField(
            model_name='complianceticket',
            name='reason',
            field=models.CharField(choices=[('trpr', 'TRPR Compliance'), ('lawn', 'LAWN Compliance'), ('paps', 'PAP Requirements'), ('afk', 'Inactivity'), ('discord_check', 'User is on discord')], max_length=20),
        ),
    ]



"""
Routines related to corporation/alliance history checks.

The functions here produce friendly HTML summaries as well as reusable
helpers that other sections (e.g. cyno readiness) use to determine how
long a member has been in corp.
"""

from datetime import timedelta
from django.utils.html import format_html
from django.utils.timezone import now, timezone
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger
from ..models import BigBrotherConfig, FrequentCorpChangesCache, CurrentStintCache
from ..app_settings import (
    ensure_datetime,
    is_npc_corporation,
    get_alliance_history_for_corp,
    get_alliance_name,
    get_corporation_info,
    get_character_employment,
)

logger = get_extension_logger(__name__)
TTL_SHORT = timedelta(hours=4)

# External site favicons, fetched each time directly from the source
ZKILL_ICON = "https://zkillboard.com/favicon.ico"
EVEWHO_ICON = "https://evewho.com/favicon.ico"
DOTLAN_ICON = "https://evemaps.dotlan.net/favicon.ico"
EVE411_ICON     = "https://www.eve411.com/favicon.ico"
FORUMS_ICON     = "https://eve-offline.net/favicon.ico"
EVESEARCH_ICON  = "https://eve-search.com/favicon.ico"


def get_frequent_corp_changes(user_id):
    """
    Build (and cache) an HTML report showing each corp membership stint.

    Hostile corps/alliances are highlighted in-line and per-character tables
    also include convenience links to the typical intel sites.
    """
    # Try 4h cache first
    try:
        cache = FrequentCorpChangesCache.objects.get(pk=user_id)
        if timezone.now() - cache.updated < TTL_SHORT:  # Serve cached card for ~4h to limit upstream calls.
            try:
                cache.last_accessed = timezone.now()
                cache.save(update_fields=["last_accessed"])
            except Exception:
                cache.save()
            return format_html(cache.html)
    except FrequentCorpChangesCache.DoesNotExist:
        pass
    # Load hostile lists
    cfg = BigBrotherConfig.get_solo()
    hostile_corps = {int(cid) for cid in cfg.hostile_corporations.split(',') if cid}  # Precompute hostile corp IDs.
    hostile_alliances = {int(aid) for aid in cfg.hostile_alliances.split(',') if aid}  # Likewise for alliances.

    characters = CharacterOwnership.objects.filter(user__id=user_id)
    html = ""

    for char in characters:
        char_name = str(char.character)
        char_id   = char.character.character_id
        try:
            history = get_character_employment(char_id)
        except Exception:
            continue

        char_links = (
            f'<a href="https://zkillboard.com/character/{char_id}/" target="_blank">'
            f'<img src="{ZKILL_ICON}" width="16" height="16" '
            f'style="margin-left:4px;vertical-align:middle;"/></a> '
            f'<a href="https://evewho.com/character/{char_id}" target="_blank">'
            f'<img src="{EVEWHO_ICON}" width="16" height="16" '
            f'style="margin-left:2px;vertical-align:middle;"/></a> '
            f'<a href="https://www.eve411.com/character/{char_id}" target="_blank">'
            f'<img src="{EVE411_ICON}" width="16" height="16" '
            f'style="margin-left:2px;vertical-align:middle;"/></a> '
            # Eve-Online forums user pages use the character name slug:
            f'<a href="https://forums.eveonline.com/u/{char_name.replace(" ", "_")}/summary" '
            f'target="_blank">'
            f'<img src="{FORUMS_ICON}" width="16" height="16" '
            f'style="margin-left:2px;vertical-align:middle;"/></a> '
            # and eve-search needs URL‐encoded name:
            f'<a href="https://eve-search.com/search/author/{char_name.replace(" ", "%20")}" '
            f'target="_blank">'
            f'<img src="{EVESEARCH_ICON}" width="16" height="16" '
            f'style="margin-left:2px;vertical-align:middle;"/></a> '
        )

        rows = []

        for idx, membership in enumerate(history):
            corp_id = membership['corporation_id']
            if is_npc_corporation(corp_id):  # Skip meaningless entries (NPC corps clutter the table).
                continue

            # Membership window
            start = ensure_datetime(membership['start_date'])
            end = ensure_datetime(history[idx+1]['start_date']) if idx+1 < len(history) else now()  # End date = next start or now.
            total_days = (end - start).days

            corp_name = get_corporation_info(corp_id)['name']
            membership_range = f"{start.date()} - {end.date()}"

            # Corp cell with external site favicons (fetched live)
            corp_color = ' class="text-danger' if (hostile_corps and corp_id in hostile_corps) else ''  # Highlight hostile corps.
            corp_cell = (
                f'<span{corp_color}>{corp_name}</span>'
                f'<a href="https://zkillboard.com/corporation/{corp_id}/" target="_blank">'
                f'<img src="{ZKILL_ICON}" width="16" height="16" style="margin-left:4px;vertical-align:middle;"/></a> '
                f'<a href="https://evewho.com/corporation/{corp_id}" target="_blank">'
                f'<img src="{EVEWHO_ICON}" width="16" height="16" style="margin-left:2px;vertical-align:middle;"/></a> '
                f'<a href="https://evemaps.dotlan.net/corp/{corp_id}" target="_blank">'
                f'<img src="{DOTLAN_ICON}" width="16" height="16" style="margin-left:2px;vertical-align:middle;"/></a> '
            )

            # Alliance segments
            alliances_html = []
            periods_html = []
            alliance_history = get_alliance_history_for_corp(corp_id)
            for j, ent in enumerate(alliance_history):
                a_start = ent['start_date']
                a_end = alliance_history[j+1]['start_date'] if j+1 < len(alliance_history) else None
                seg_start = max(start, a_start)
                seg_end = min(end, a_end) if a_end else end
                if seg_start < seg_end:  # Only render overlapping time periods (ignore non-overlaps).
                    aid = ent['alliance_id']
                    aname = get_alliance_name(aid)
                    period = f"{seg_start.date()} - {seg_end.date()}"

                    if aid:  # Only render alliance rows when the corp was in an alliance.
                        alliance_color = ' class="text-danger"' if (hostile_alliances and aid in hostile_alliances) else ''  # Flag hostile alliances.
                        name_cell = f'<span{alliance_color}>{aname}</span>'
                        icons = (
                            f'<a href="https://zkillboard.com/alliance/{aid}/" target="_blank">'
                            f'<img src="{ZKILL_ICON}" width="16" height="16" style="margin-left:4px;vertical-align:middle;"/></a> '
                            f'<a href="https://evewho.com/alliance/{aid}" target="_blank">'
                            f'<img src="{EVEWHO_ICON}" width="16" height="16" style="margin-left:2px;vertical-align:middle;"/></a> '
                            f'<a href="https://evemaps.dotlan.net/alliance/{aid}" target="_blank">'
                            f'<img src="{DOTLAN_ICON}" width="16" height="16" style="margin-left:2px;vertical-align:middle;"/></a> '
                        )
                    else:
                        name_cell = '-'
                        icons = ''
                    alliances_html.append(name_cell + icons)
                    periods_html.append(period)

            if not alliances_html:  # When no alliance data, fallback to corp membership range.
                alliances_html = ['-']
                periods_html = [membership_range]

            # Duration cell coloring only
            dur_color = ' class="text-danger"' if total_days < 10 else (' class="text-warning"' if total_days < 30 else '')  # Quick visual for recent corps.

            rows.append({
                'corp_cell': corp_cell,
                'membership_range': membership_range,
                'alliances_html': '<br>'.join(alliances_html),
                'periods_html': '<br>'.join(periods_html),
                'total_days': total_days,
                'dur_color': dur_color,
            })

        html += format_html("<h3>{} {}</h3>", char_name, format_html(char_links))
        html += '<table class="table table-striped table-hover stats">'
        html += '<thead><tr><th>Corporation</th><th>Membership</th><th>Alliance(s)</th><th>Alliance Dates</th><th>Time in Corp</th></tr></thead><tbody>'
        for r in rows:
            row_html = (
                '<tr>'
                f'<td>{r["corp_cell"]}</td>'
                f'<td>{r["membership_range"]}</td>'
                f'<td>{r["alliances_html"]}</td>'
                f'<td>{r["periods_html"]}</td>'
                f'<td{r["dur_color"]}>{r["total_days"]} days</td>'
                '</tr>'
            )
            html += format_html(row_html)
        html += '</tbody></table>'

    # Save cache
    try:
        FrequentCorpChangesCache.objects.update_or_create(
            user_id=user_id, defaults={"html": str(html), "last_accessed": timezone.now()}
        )
    except Exception:
        pass
    return format_html(html)

def time_in_corp(user_id):
    """
    Return the maximum number of days any of the user's characters have been
    continuously in the configured main corporation.
    """
    days = 0
    characters = CharacterOwnership.objects.filter(user__id=user_id)
    for char in characters:
        char_id   = char.character.character_id
        c_days = get_current_stint_days_in_corp(char_id, BigBrotherConfig.get_solo().main_corporation_id)
        if c_days > days:  # Track the maximum stint across all characters.
            days = c_days
    return days

def get_current_stint_days_in_corp(char_id: int, corp_id: int) -> int:
    """
    Return the number of days the character has been in the given corporation *currently*.
    If the character is not in that corporation right now, return 0.
    """
    try:
        if is_npc_corporation(corp_id):  # NPC corps don't matter for stint tracking.
            return 0

        # 4h cached value
        try:
            cache = CurrentStintCache.objects.get(char_id=char_id, corp_id=corp_id)
            if timezone.now() - cache.updated < TTL_SHORT:  # Serve cached days when fresh.
                try:
                    cache.last_accessed = timezone.now()
                    cache.save(update_fields=["last_accessed"])
                except Exception:
                    cache.save()
                return int(cache.days)
        except CurrentStintCache.DoesNotExist:
            pass

        history = get_character_employment(char_id)

        if not history:  # No employment history returned -> treat as zero.
            return 0

        # Latest membership entry appears at the end of the history list
        latest = history[-1]
        if latest["corporation_id"] != corp_id:  # Character left the corp already.
            return 0

        start = ensure_datetime(latest["start_date"])
        end = now()

        days = max(0, (end - start).days)
        try:
            CurrentStintCache.objects.update_or_create(
                char_id=char_id,
                corp_id=corp_id,
                defaults={"days": days, "last_accessed": timezone.now()},
            )
        except Exception:
            pass
        return days

    except Exception as e:
        logger.exception(
            "Failed to compute current stint days for char %s in corp %s: %s",
            char_id, corp_id, e
        )
        return 0



"""
Django signal handlers used by BigBrother.

Currently:
1. When the singleton config is saved, Celery message tasks stay in sync.
2. When a character ownership is deleted, optionally open a compliance ticket.
"""

from django.dispatch import receiver
from django.db.models.signals import post_save, pre_delete

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

from .models import BigBrotherConfig, TicketToolConfig
from .tasks_cb import BB_register_message_tasks
from .tasks_tickets import get_webhook_for_reason
from .app_settings import send_message, send_status_embed


logger = get_extension_logger(__name__)

try:
    from aadiscordbot.tasks import run_task_function
    from aadiscordbot.utils.auth import get_discord_user_id
except ImportError:
    run_task_function = None
    get_discord_user_id = None
    logger.info("✅  [AA-BB] - [Signals] - aadiscordbot not installed, signaling will use fallbacks.")

@receiver(post_save, sender=BigBrotherConfig)
@receiver(post_save, sender=TicketToolConfig)
def trigger_task_sync(sender, instance, **kwargs):
    """When the config changes, make sure Celery schedules match the DB."""
    BB_register_message_tasks.delay()


@receiver(pre_delete, sender=CharacterOwnership)
def removed_character(sender, instance, **kwargs):
    """
    If the ticket tool is monitoring “character removed” events, raise a ticket
    any time Auth loses access to one of the pilot’s characters.
    """
    if not TicketToolConfig.get_solo().char_removed_enabled:
        return
    try:
        character = instance.character
        bb_cfg = BigBrotherConfig.get_solo()
        member_states = bb_cfg.bb_member_states.all()
        if instance.user.profile.state not in member_states:
            return

        if bb_cfg.limit_to_main_corp:
            # Check if the user's main character belongs to the primary corporation
            profile = getattr(instance.user, 'profile', None)
            main_char = getattr(profile, 'main_character', None) if profile else None
            if not main_char or main_char.corporation_id != bb_cfg.main_corporation_id:
                return

        from .tasks_tickets import ensure_ticket
        ensure_ticket(instance.user, "char_removed", details=str(character))

    except Exception as e:
        logger.error("✅  [AA-BB] - [Signals] - Failed to create character-removed ticket: %s", e)



from allianceauth.services.hooks import get_extension_logger

from django.db import transaction, IntegrityError
from django_celery_beat.models import PeriodicTask, CrontabSchedule, IntervalSchedule

logger = get_extension_logger(__name__)

def format_task_name(name: str) -> str:
    """
    Standardize a task name with proper prefix and capitalization.
    'BB: ' is used for BB tasks, 'CB: ' for CB tasks.
    """
    # 1. Strip any existing known prefixes
    prefixes = ["AA-BB: ", "BB: ", "CB: "]
    raw_name = name
    original_prefix = None
    for p in prefixes:
        if raw_name.startswith(p):
            original_prefix = p
            raw_name = raw_name[len(p):]
            break

    # 2. Determine the new prefix and strip "BB " or "CB " if it's there
    if raw_name.upper().startswith("BB "):
        actual_prefix = "BB: "
        raw_name = raw_name[3:]
    elif raw_name.upper().startswith("CB "):
        actual_prefix = "CB: "
        raw_name = raw_name[3:]
    elif original_prefix == "CB: ":
        actual_prefix = "CB: "
    elif raw_name.lower().startswith("tickets run"):
        actual_prefix = "BB: "
    else:
        actual_prefix = "BB: "

    def format_word(word):
        upper_word = word.upper()
        if upper_word in ["BB", "CB", "CT", "DB", "AA", "ESI", "EVE", "API"]:
            return upper_word
        elif upper_word == "LOA":
            return "LoA"
        return word.capitalize()

    # Replace hyphens with spaces for better capitalization
    raw_name = raw_name.replace('-', ' ')
    formatted_words = [format_word(w) for w in raw_name.split()]
    return actual_prefix + " ".join(formatted_words)


def setup_periodic_task(
    name: str,
    task_path: str,
    schedule,
    enabled: bool = False,
    update_schedule: bool = False,
):
    """
    Create or update a periodic task consistently.
    Ensures naming scheme 'BB: ' or 'CB: ' and proper capitalization.
    Renames existing tasks if necessary.
    Never deletes tasks.
    """
    standardized_name = format_task_name(name)

    # 2. Find existing task
    task = PeriodicTask.objects.filter(name=standardized_name).first()
    old_task = PeriodicTask.objects.filter(name=name).first() if standardized_name != name else None

    if task and old_task and task.pk != old_task.pk:
        # Both exist! Delete the old one to avoid duplicates.
        logger.info(f"Found both '{standardized_name}' and '{name}'. Deleting the old one.")
        old_task.delete()
    elif not task and old_task:
        # Only old one exists, rename it
        task = old_task
        logger.info(f"Renaming periodic task '{name}' to '{standardized_name}'")
        task.name = standardized_name

    # 3. Create or update
    updated = False
    is_new = False
    if not task:
        task = PeriodicTask(name=standardized_name)
        updated = True
        is_new = True

    if task.task != task_path:
        task.task = task_path
        updated = True

    if is_new or update_schedule:
        if isinstance(schedule, CrontabSchedule):
            if task.crontab != schedule:
                task.crontab = schedule
                task.interval = None
                updated = True
        elif isinstance(schedule, IntervalSchedule):
            if task.interval != schedule:
                task.interval = schedule
                task.crontab = None
                updated = True

    if task.enabled != enabled:
        task.enabled = enabled
        updated = True

    if updated:
        try:
            with transaction.atomic():
                task.save()
            logger.info(f"✅  [AA-BB] - [Tasks] - {'Created' if task.pk is None else 'Updated'} '{standardized_name}' periodic task (enabled={enabled})")
        except IntegrityError:
            # Handle race condition where another process might have created it
            logger.warning(f"IntegrityError while saving periodic task '{standardized_name}', fetching existing.")
            task = PeriodicTask.objects.get(name=standardized_name)
    else:
        logger.info(f"ℹ️  [AA-BB] - [Tasks] - '{standardized_name}' periodic task already exists and is up to date")

    return task


def sync_periodic_tasks():
    """
    Consolidated function to ensure all BigBrother periodic tasks exist and are properly configured.
    Respects feature toggles and syncs schedules from configuration.
    """
    from .models import BigBrotherConfig
    from django.apps import apps

    config = BigBrotherConfig.get_solo()
    config.refresh_from_db()
    master_active = config.is_active
    logger.info(f"ℹ️  [AA-BB] - [Tasks] - Starting sync_periodic_tasks. Master Active: {master_active}")

    # --- Schedules ---

    # Standard update interval (dynamic based on stagger window)
    stagger = max(getattr(config, "update_stagger_seconds", 3600), 3600)
    interval, _ = IntervalSchedule.objects.get_or_create(
        every=stagger,
        period=IntervalSchedule.SECONDS,
    )

    # Hourly at :25 past
    hourly_25_schedule, _ = CrontabSchedule.objects.get_or_create(
        minute='25',
        hour='*',
        day_of_week='*',
        day_of_month='*',
        month_of_year='*',
        timezone='UTC',
    )

    # Daily at 12:00 UTC
    daily_12_schedule, _ = CrontabSchedule.objects.get_or_create(
        minute='0',
        hour='12',
        day_of_week='*',
        day_of_month='*',
        month_of_year='*',
        timezone='UTC',
    )

    # Weekly on Sunday at 12:00 UTC
    weekly_sun_12_schedule, _ = CrontabSchedule.objects.get_or_create(
        minute='0',
        hour='12',
        day_of_week='0',
        day_of_month='*',
        month_of_year='*',
        timezone='UTC',
    )

    # Daily cleanup at 01:00 UTC
    cleanup_schedule, _ = CrontabSchedule.objects.get_or_create(
        minute='0',
        hour='1',
        day_of_week='*',
        day_of_month='*',
        month_of_year='*',
        timezone='UTC',
    )

    # --- Task List ---
    # format: name, task_path, schedule, active_attr, update_schedule, is_master_dependent
    # if schedule_attr is provided, it overrides schedule from config
    tasks_to_sync = [
        {
            "name": "BB run regular updates",
            "task_path": "aa_bb.tasks.BB_run_regular_updates",
            "schedule": interval,
            "active_attr": "is_active",
            "update_schedule": True,
        },
        {
            "name": "CB run regular updates",
            "task_path": "aa_bb.tasks_cb.CB_run_regular_updates",
            "schedule": hourly_25_schedule,
            "active_attr": "is_active",
            "update_schedule": True,
        },
        {
            "name": "BB kickstart stale CT modules",
            "task_path": "aa_bb.tasks_ct.kickstart_stale_ct_modules",
            "schedule": hourly_25_schedule,
            "active_attr": "is_active",
            "update_schedule": True,
        },
        {
            "name": "BB run regular DB cleanup",
            "task_path": "aa_bb.tasks_cb.BB_daily_DB_cleanup",
            "schedule": cleanup_schedule,
            "active_attr": "is_active",
            "update_schedule": True,
        },
        {
            "name": "tickets run regular updates",
            "task_path": "aa_bb.tasks_tickets.hourly_compliance_check",
            "schedule": hourly_25_schedule,
            "active_attr": "is_active", # We use is_active as the master switch for tickets too
            "update_schedule": True,
        },
        {
            "name": "BB run regular LoA updates",
            "task_path": "aa_bb.tasks_cb.BB_run_regular_loa_updates",
            "schedule": daily_12_schedule,
            "active_attr": "is_loa_active",
            "update_schedule": True,
        },
        {
            "name": "BB check member compliance",
            "task_path": "aa_bb.tasks_cb.check_member_compliance",
            "schedule": daily_12_schedule,
            "active_attr": "is_paps_active",
            "update_schedule": True,
        },
        {
            "name": "BB send recurring stats",
            "task_path": "aa_bb.tasks_other.BB_send_recurring_stats",
            "schedule_attr": "stats_schedule",
            "default_schedule": weekly_sun_12_schedule,
            "active_attr": "are_recurring_stats_active",
            "update_schedule": True,
        },
        {
            "name": "BB send daily message",
            "task_path": "aa_bb.tasks_cb.BB_send_daily_messages",
            "schedule_attr": "dailyschedule",
            "default_schedule": daily_12_schedule,
            "active_attr": "are_daily_messages_active",
            "update_schedule": True,
        },
        {
            "name": "BB send optional message 1",
            "task_path": "aa_bb.tasks_cb.BB_send_opt_message1",
            "schedule_attr": "optschedule1",
            "default_schedule": daily_12_schedule,
            "active_attr": "are_opt_messages1_active",
            "update_schedule": True,
        },
        {
            "name": "BB send optional message 2",
            "task_path": "aa_bb.tasks_cb.BB_send_opt_message2",
            "schedule_attr": "optschedule2",
            "default_schedule": daily_12_schedule,
            "active_attr": "are_opt_messages2_active",
            "update_schedule": True,
        },
        {
            "name": "BB send optional message 3",
            "task_path": "aa_bb.tasks_cb.BB_send_opt_message3",
            "schedule_attr": "optschedule3",
            "default_schedule": daily_12_schedule,
            "active_attr": "are_opt_messages3_active",
            "update_schedule": True,
        },
        {
            "name": "BB send optional message 4",
            "task_path": "aa_bb.tasks_cb.BB_send_opt_message4",
            "schedule_attr": "optschedule4",
            "default_schedule": daily_12_schedule,
            "active_attr": "are_opt_messages4_active",
            "update_schedule": True,
        },
        {
            "name": "BB send optional message 5",
            "task_path": "aa_bb.tasks_cb.BB_send_opt_message5",
            "schedule_attr": "optschedule5",
            "default_schedule": daily_12_schedule,
            "active_attr": "are_opt_messages5_active",
            "update_schedule": True,
        },
    ]

    # Add optional tasks if dependencies are met
    if apps.is_installed("aa_contacts"):
        tasks_to_sync.append({
            "name": "BB sync contacts from aa-contacts",
            "task_path": "aa_bb.tasks.BB_sync_contacts_from_aa_contacts",
            "schedule": hourly_25_schedule,
            "active_attr": "auto_import_contacts_enabled",
            "update_schedule": True,
        })

    for task_info in tasks_to_sync:
        name = task_info["name"]
        task_path = task_info["task_path"]

        # Resolve schedule
        schedule_attr = task_info.get("schedule_attr")
        if schedule_attr:
            schedule = getattr(config, schedule_attr, None) or task_info.get("default_schedule") or daily_12_schedule
        else:
            schedule = task_info.get("schedule")

        # Resolve enabled status (master switch + feature toggle)
        feature_enabled = bool(getattr(config, task_info["active_attr"], False))
        enabled = master_active and feature_enabled

        # Special case: 'is_active' tasks ARE the master switch
        if task_info["active_attr"] == "is_active":
            enabled = master_active

        logger.debug(f"ℹ️  [AA-BB] - [Tasks] - Task '{name}': feature_enabled={feature_enabled}, master_active={master_active} -> enabled={enabled}")

        # Log deactivation reason if it's currently enabled but about to be disabled
        from django_celery_beat.models import PeriodicTask
        existing_task = PeriodicTask.objects.filter(name=format_task_name(name)).first()
        if existing_task and existing_task.enabled and not enabled:
            if not master_active:
                logger.warning(f"ℹ️  [AA-BB] - [Tasks] - '{name}' will be DISABLED because Big Brother is globally inactive.")
            elif not feature_enabled:
                logger.warning(f"ℹ️  [AA-BB] - [Tasks] - '{name}' will be DISABLED because its feature toggle '{task_info['active_attr']}' is False.")

        setup_periodic_task(
            name=name,
            task_path=task_path,
            schedule=schedule,
            enabled=enabled,
            update_schedule=task_info.get("update_schedule", False)
        )



# Generated by Django 4.2.23 on 2025-08-05 09:32

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('aa_bb', '0025_alter_messages_options_alter_optmessages1_options_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False, 'permissions': (('basic_access', 'Can access this app'), ('full_access', 'Can view all main characters'), ('recruiter_access', 'Can view guest main characters only'), ('can_blacklist_characters', 'Can add characters to blacklist'), ('can_access_loa', 'Can submit a Leave Of Absence request'), ('can_manage_loa', 'Can manage Leave Of Absence requests'))},
        ),
        migrations.AlterField(
            model_name='bigbrotherconfig',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='entityinfocache',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='messages',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages1',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages2',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages3',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages4',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='optmessages5',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='suscontractnote',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='susmailnote',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='sustransactionnote',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='userstatus',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='warmprogress',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.CreateModel(
            name='LeaveRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_date', models.DateField()),
                ('end_date', models.DateField()),
                ('reason', models.TextField()),
                ('status', models.CharField(choices=[('pending', 'Pending'), ('approved', 'Approved'), ('denied', 'Denied')], default='pending', max_length=20)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='leave_requests', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# Generated by Django 4.2.23 on 2025-08-18 17:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0048_characteraccountstate'),
    ]

    operations = [
        migrations.AddField(
            model_name='userstatus',
            name='clone_status',
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



"""
Wallet transaction hygiene checks. These helpers normalize journal rows,
flag suspicious counterparties, and keep deduplicated notes for alerts.
"""

import html
import requests
from typing import Dict, Optional
from datetime import datetime, timedelta
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)

from ..app_settings import (
    get_user_characters,
    get_entity_info,
    get_safe_entities,
    aablacklist_active,
    resolve_location_name,
    resolve_location_system_id,
    is_location_hostile,
    get_system_owner,
    get_hostile_state,
    is_highsec,
    is_lowsec,
    corptools_active,
    is_hostile_unified,
)

if aablacklist_active():
    from .add_to_blacklist import check_char_add_to_bl
else:
    def check_char_corp_bl(_cid: int) -> bool:
        return False

try:
    if corptools_active():
        from corptools.models import (
            CharacterWalletJournalEntry as WalletJournalEntry,
            Structure,
        )
        # Try to import CharacterMarketTransaction, but don't fail if it doesn't exist
        try:
            from corptools.models import CharacterMarketTransaction
        except ImportError:
            CharacterMarketTransaction = None
            logger.warning("CharacterMarketTransaction not available in corptools")

        logger.info(f"Successfully imported WalletJournalEntry: {WalletJournalEntry}")
    else:
        logger.warning("corptools_active() returned False at import time")
        WalletJournalEntry = None
        CharacterMarketTransaction = None
        Structure = None
except ImportError as e:
    logger.error(f"Failed to import corptools models: {e}")
    WalletJournalEntry = None
    CharacterMarketTransaction = None
    Structure = None

from django.apps import apps
EVEUNIVERSE_INSTALLED = apps.is_installed("eveuniverse")
if EVEUNIVERSE_INSTALLED:
    try:
        from eveuniverse.models import EveMarketPrice
    except ImportError:
        EVEUNIVERSE_INSTALLED = False

from ..models import BigBrotherConfig, ProcessedTransaction, SusTransactionNote, EveItemPrice

SUS_TYPES = ("player_trading", "corporation_account_withdrawal", "player_donation")


def _find_employment_at(employment: list, date: datetime) -> Optional[dict]:
    for rec in employment:
        start = rec.get("start_date")
        end = rec.get("end_date")
        if start and start <= date and (end is None or date < end):
            return rec
    return None


def _find_alliance_at(history: list, date: datetime) -> Optional[int]:
    for i, rec in enumerate(history):
        start = rec.get("start_date")
        next_start = history[i + 1]["start_date"] if i + 1 < len(history) else None
        if start and start <= date and (next_start is None or date < next_start):
            return rec.get("alliance_id")
    return None


def gather_user_transactions(user_id: int):
    if not corptools_active() or WalletJournalEntry is None:
        return []

    user_chars = get_user_characters(user_id)
    user_ids = set(user_chars.keys())
    logger.info(f"gather_user_transactions for user {user_id}: user_chars={list(user_ids)}")

    # Check total entries in database
    total_entries = WalletJournalEntry.objects.count()
    logger.info(f"Total CharacterWalletJournalEntry records in database: {total_entries}")

    # Check entries for user's characters
    char_entries = WalletJournalEntry.objects.filter(character__character__character_id__in=user_ids)
    logger.info(f"Entries for user's characters: {char_entries.count()}")

    if char_entries.exists():
        sample = char_entries.first()
        logger.info(f"Sample entry: character_id={sample.character.character.character_id}, first_party={sample.first_party_id}, second_party={sample.second_party_id}, ref_type={sample.ref_type}")

    from django.db.models import Q
    # Filter by character ownership (entries belonging to user's characters)
    qs = WalletJournalEntry.objects.filter(character__character__character_id__in=user_ids)

    # Also filter to transactions involving external parties
    # Keep only transactions where at least one party is NOT the user
    qs = qs.filter(Q(first_party_id__in=user_ids) | Q(second_party_id__in=user_ids))
    logger.info(f"After filtering by first_party_id or second_party_id in user_ids: {qs.count()}")

    qs = qs.exclude(first_party_id__in=user_ids, second_party_id__in=user_ids)
    logger.info(f"After excluding internal transactions: {qs.count()}")

    return qs


def get_user_transactions(qs) -> Dict[int, Dict]:
    result: Dict[int, Dict] = {}

    _info_cache: dict[tuple[int, int], dict] = {}

    def _cached_info(eid: int, when: datetime) -> dict:
        key = (int(eid or 0), int(when.date().toordinal()))
        if key in _info_cache:
            return _info_cache[key]
        info = get_entity_info(eid, when)
        _info_cache[key] = info
        return info

    for entry in qs:
        tx_id = entry.entry_id
        tx_date = entry.date

        first_party_id = entry.first_party_id
        iinfo = _cached_info(first_party_id, tx_date)

        second_party_id = entry.second_party_id
        ainfo = _cached_info(second_party_id, tx_date)

        context_id = entry.context_id
        context_type = entry.context_id_type
        system_id = None
        location_id = None
        type_id = None
        quantity = 1
        if context_type == "structure_id":
            name = resolve_location_name(context_id)
            context = f"Structure: {name}" if name else f"Structure ID: {context_id}"
            location_id = context_id
            system_id = resolve_location_system_id(context_id)
        elif context_type == "character_id":
            context = f"Character: {_cached_info(context_id, tx_date)['name']}"
        elif context_type == "eve_system":
            context = "EVE System"
            system_id = context_id
            location_id = context_id
        elif context_type is None:
            context = "None"
        elif context_type == "market_transaction_id":
            context = f"Market Transaction ID: {context_id}"
            if CharacterMarketTransaction:
                m_tx = CharacterMarketTransaction.objects.filter(transaction_id=context_id).first()
                if m_tx:
                    location_id = m_tx.location_id if hasattr(m_tx, "location_id") else None
                    system_id = m_tx.location.system_id if hasattr(m_tx.location, "system_id") else None
                    type_id = m_tx.type_id
                    quantity = m_tx.quantity
        else:
            context = f"{context_type}: {context_id}"

        result[tx_id] = {
            "entry_id": tx_id,
            "date": tx_date,
            "amount": "{:,}".format(entry.amount),
            "raw_amount": float(entry.amount),
            "balance": "{:,}".format(entry.balance),
            "description": entry.description,
            "reason": entry.reason,
            "first_party_id": first_party_id,
            "first_party_name": iinfo["name"],
            "first_party_corporation_id": iinfo["corp_id"],
            "first_party_corporation": iinfo["corp_name"],
            "first_party_alliance_id": iinfo["alli_id"],
            "first_party_alliance": iinfo["alli_name"],
            "second_party_id": second_party_id,
            "second_party_name": ainfo["name"],
            "second_party_corporation_id": ainfo["corp_id"],
            "second_party_corporation": ainfo["corp_name"],
            "second_party_alliance_id": ainfo["alli_id"],
            "second_party_alliance": ainfo["alli_name"],
            "context": context,
            "type": entry.ref_type,
            "system_id": system_id,
            "location_id": location_id,
            "type_id": type_id,
            "quantity": quantity,
        }

    return result


def is_transaction_hostile(tx: dict, user_ids: set = None, safe_entities: set = None) -> bool:
    """
    Checks if a wallet transaction is considered hostile using the unified processor.
    """
    ttype = tx.get("type") or ""
    is_sus_type = any(st in ttype for st in SUS_TYPES)
    is_market = "market_escrow" in ttype or "market_transaction" in ttype

    if not (is_sus_type or is_market):
        return False

    fpid = tx.get("first_party_id")
    spid = tx.get("second_party_id")

    # Unified check handles Rule 1 (Safe entities), which includes both parties being safe.
    # It also handles hubs, price thresholds, location hostility, etc.
    return is_hostile_unified(
        involved_ids=[fpid, spid],
        location_id=tx.get("location_id"),
        system_id=tx.get("system_id"),
        is_market=is_market,
        market_item_id=tx.get("type_id"),
        market_unit_price=abs(tx.get("raw_amount")) / (tx.get("quantity") or 1) if tx.get("raw_amount") is not None else None,
        when=tx.get("date"),
        safe_entities=safe_entities
    )


def render_transactions(user_id: int) -> str:
    """
    Render HTML table of recent hostile wallet transactions for user
    """
    qs = gather_user_transactions(user_id)
    txs = get_user_transactions(qs)

    user_chars = get_user_characters(user_id)
    user_ids = set(user_chars.keys())

    from ..app_settings import get_safe_entities
    safe_entities = get_safe_entities()

    # sort by date desc
    all_list = sorted(txs.values(), key=lambda x: x['date'], reverse=True)
    hostile = [t for t in all_list if is_transaction_hostile(t, user_ids, safe_entities=safe_entities)]
    if not hostile:  # No transactions require attention.
        return '<p>No hostile transactions found.</p>'

    limit = 50
    display = hostile[:limit]
    skipped = max(0, len(hostile) - limit)

    # define headers to show
    first = display[0]
    HIDDEN = {'first_party_id','second_party_id','first_party_corporation_id','second_party_corporation_id',
              'first_party_alliance_id','second_party_alliance_id','entry_id'}
    headers = [k for k in first.keys() if k not in HIDDEN]

    parts = ['<table class="table table-striped table-hover stats">','<thead>','<tr>']
    for h in headers:
        parts.append(f'<th>{html.escape(h.replace("_"," ").title())}</th>')
    parts.extend(['</tr>','</thead>','<tbody>'])

    cfg = BigBrotherConfig.get_solo()
    hostile_corps = {s.strip() for s in (cfg.hostile_corporations or "").split(",") if s.strip()}
    hostile_allis = {s.strip() for s in (cfg.hostile_alliances or "").split(",") if s.strip()}
    safe_entities = get_safe_entities()

    for t in display:  # Render each hostile transaction row with contextual styling.
        parts.append('<tr>')
        for col in headers:
            val = html.escape(str(t.get(col)))
            style = ''
            # reuse contract style logic by mapping to transaction
            if col == 'type':
                for key in SUS_TYPES:
                    if key in t['type']:  # Highlight suspicious ref types inline.
                        style = 'color: red;'
                if cfg.show_market_transactions:
                    if "market_escrow" in t['type'] or "market_transaction" in t['type']:
                        style = 'color: red;'
            if col in ('first_party_name', 'second_party_name'):
                pid = t.get(col + '_id')
                if get_hostile_state(pid, 'character'):
                    style = 'color: red;'
            if col.endswith('corporation'):
                cid = t.get(col + '_id')
                if get_hostile_state(cid, 'corporation'):
                    style = 'color: red;'
            if col.endswith('alliance'):
                aid = t.get(col + '_id')
                if get_hostile_state(aid, 'alliance'):
                    style = 'color: red;'
            def make_td(val, style=""):
                """Render a TD with optional inline style for hostile cues."""
                style_attr = f' style="{style}"' if style else ""
                return f"<td{style_attr}>{val}</td>"
            parts.append(make_td(val, style))
        parts.append('</tr>')

    parts.extend(['</tbody>','</table>'])
    if skipped:  # Let the reviewer know older hostile rows are omitted.
        parts.append(f'<p>Showing {limit} of {len(hostile)} hostile transactions; skipped {skipped} older ones.</p>')
    return '\n'.join(parts)


def get_user_hostile_transactions(user_id: int) -> Dict[int, str]:
    qs_all = gather_user_transactions(user_id)
    all_ids = list(qs_all.values_list("entry_id", flat=True))

    seen = set(
        ProcessedTransaction.objects.filter(entry_id__in=all_ids).values_list("entry_id", flat=True)
    )

    notes: Dict[int, str] = {}
    new = [eid for eid in all_ids if eid not in seen]

    if new:
        new_qs = qs_all.filter(entry_id__in=new)
        rows = get_user_transactions(new_qs)

        user_chars = get_user_characters(user_id)
        user_ids = set(user_chars.keys())

        from ..app_settings import get_safe_entities
        safe_entities = get_safe_entities()

        hostile_rows: dict[int, dict] = {eid: tx for eid, tx in rows.items() if is_transaction_hostile(tx, user_ids, safe_entities=safe_entities)}
        if hostile_rows:
            ProcessedTransaction.objects.bulk_create(
                [ProcessedTransaction(entry_id=eid) for eid in hostile_rows.keys()],
                ignore_conflicts=True,
            )
            pts = {
                pt.entry_id: pt
                for pt in ProcessedTransaction.objects.filter(entry_id__in=hostile_rows.keys())
            }

            cfg = BigBrotherConfig.get_solo()
            show_market = cfg.show_market_transactions

            for eid, tx in hostile_rows.items():
                pt = pts.get(eid)
                if not pt:
                    continue

                flags = []
                ttype = tx.get("type") or ""
                for key in SUS_TYPES:
                    if key in ttype:
                        flags.append(f"Transaction type is **{ttype}**")

                if show_market:
                    if "market_escrow" in ttype or "market_transaction" in ttype:
                        flags.append(f"Transaction type is **{ttype}**")

                fpid = tx.get("first_party_id")
                if get_hostile_state(fpid, 'character'):
                    flags.append(f"first_party **{tx['first_party_name']}** is hostile/blacklisted")

                spid = tx.get("second_party_id")
                if get_hostile_state(spid, 'character'):
                    flags.append(f"second_party **{tx['second_party_name']}** is hostile/blacklisted")

                loc_id = tx.get("location_id") or tx.get("system_id")
                if loc_id and is_location_hostile(tx.get("location_id"), tx.get("system_id")):
                    loc_name = resolve_location_name(loc_id) or f"ID {loc_id}"
                    owner_info = get_system_owner({"id": loc_id})
                    oname = owner_info.get("owner_name")
                    rname = owner_info.get("region_name")
                    flag = f"Location **{loc_name}** is hostile space"
                    if oname or rname:
                        info_parts = []
                        if oname:
                            info_parts.append(oname)
                        if rname and rname != "Unknown Region":
                            info_parts.append(f"Region: {rname}")
                        flag += f" ({' | '.join(info_parts)})"
                    flags.append(flag)

                flags_lines = [f"    - {flag}" for flag in flags] if flags else ["    - (no extra flags)"]

                note_lines = [
                    f"- **{tx['date']}** · **{tx['amount']} ISK**",
                    f"  - **Type:** {tx['type']}",
                    (
                        f"  - **From:** {tx['first_party_name']} "
                        f"({tx['first_party_corporation']} | {tx['first_party_alliance']})"
                    ),
                    (
                        f"  - **To:** {tx['second_party_name']} "
                        f"({tx['second_party_corporation']} | {tx['second_party_alliance']})"
                    ),
                ]

                if tx.get("reason") and tx.get("reason") != "None":
                    note_lines.append(f"  - **Reason:** {tx['reason']}")

                if tx.get("context") and tx.get("context") != "None":
                    note_lines.append(f"  - **Context:** {tx['context']}")

                note_lines.append("  - **Flags:**")
                note_lines.extend(flags_lines)

                note = "\n".join(note_lines)

                SusTransactionNote.objects.update_or_create(
                    transaction=pt,
                    defaults={"user_id": user_id, "note": note},
                )
                notes[eid] = note

    for note_obj in SusTransactionNote.objects.filter(user_id=user_id):
        notes[note_obj.transaction.entry_id] = note_obj.note

    return notes



"""Shared configuration for the Django-ESI OpenAPI 3.1 client."""

from __future__ import annotations

from django.utils import timezone
from email.utils import parsedate_to_datetime
from esi.openapi_clients import ESIClientProvider

from . import __title__, __version__, __github_url__, __esi_compatibility_date__

DEFAULT_OPERATIONS = [
    "PostUniverseNames",
    "PostUniverseIds",
    "GetCharactersCharacterIdCorporationhistory",
    "GetCorporationsCorporationId",
    "GetCorporationsCorporationIdAlliancehistory",
    "GetAlliancesAllianceId",
    "GetSovereigntyMap",
    "GetKillmailsKillmailIdKillmailHash",
]


esi = ESIClientProvider(
    compatibility_date=__esi_compatibility_date__,
    ua_appname=__title__,
    ua_version=__version__,
    ua_url=__github_url__,
    operations=DEFAULT_OPERATIONS,
)


def to_plain(value):
    """Recursively convert Pydantic models returned by the OpenAPI client to plain Python types."""
    if hasattr(value, "model_dump"):
        return to_plain(value.model_dump())
    if isinstance(value, list):
        return [to_plain(item) for item in value]
    if isinstance(value, dict):
        return {key: to_plain(val) for key, val in value.items()}
    return value


def parse_expires(headers: dict | None):
    """Extract a timezone-aware datetime from HTTP Expires headers (if present)."""
    if not headers:
        return None
    value = headers.get("Expires")
    if not value:
        return None
    try:
        dt = parsedate_to_datetime(value)
    except (TypeError, ValueError):
        return None
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    return dt.astimezone(timezone.utc)


def call_result(operation, **kwargs):
    """Execute an OpenAPI operation.result() call and return (data, expires_at)."""
    data, response = operation.result(return_response=True, **kwargs)
    return to_plain(data), parse_expires(response.headers)


def call_results(operation, **kwargs):
    """Execute operation.results() and return (list_data, expires_at) with plain types."""
    data, response = operation.results(return_response=True, **kwargs)
    return to_plain(data), parse_expires(response.headers)



# Generated by Django 4.2.20 on 2025-05-09 04:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('aa_bb', '0016_warmprogress'),
    ]

    operations = [
        migrations.CreateModel(
            name='EntityInfoCache',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('entity_id', models.IntegerField()),
                ('as_of', models.DateTimeField()),
                ('data', models.JSONField()),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
            options={
                'indexes': [models.Index(fields=['entity_id', 'as_of'], name='aa_bb_entit_entity__79afd6_idx'), models.Index(fields=['updated'], name='aa_bb_entit_updated_29e0d0_idx')],
                'unique_together': {('entity_id', 'as_of')},
            },
        ),
    ]
